#
# Copyright (c) 2023 Nordic Semiconductor ASA
#
# SPDX-License-Identifier: Apache-2.0
#

"""
Various internal functionality used by the device module.
"""

from __future__ import annotations

import functools as ft
import re
import typing
from abc import ABC
from collections import defaultdict
from time import perf_counter_ns
from types import MappingProxyType
from typing import (
    Any,
    Callable,
    Collection,
    Dict,
    Generic,
    Iterable,
    Iterator,
    List,
    Mapping,
    Optional,
    Protocol,
    Reversible,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
)

import lxml.etree as ET

from . import bindings
from .path import EPath


def topo_sort_derived_peripherals(
    peripherals: Iterable[bindings.PeripheralElement],
) -> List[bindings.PeripheralElement]:
    """
    Topologically sort the peripherals based on 'derivedFrom' attributes using Kahn's algorithm.
    The returned list has the property that the peripheral element at index i does not derive from
    any of the peripherals at indices 0..(i - 1).

    :param peripherals: List of peripheral elements to sort
    :return: List of peripheral elements topologically sorted based on the 'derivedFrom' attribute.
    """

    sorted_peripherals: List[bindings.PeripheralElement] = []
    no_dep_peripherals: List[bindings.PeripheralElement] = []
    dep_graph: Dict[str, List[bindings.PeripheralElement]] = defaultdict(list)

    for peripheral in peripherals:
        if peripheral.derived_from is not None:
            dep_graph[peripheral.derived_from].append(peripheral)
        else:
            no_dep_peripherals.append(peripheral)

    while no_dep_peripherals:
        peripheral = no_dep_peripherals.pop()
        sorted_peripherals.append(peripheral)
        # Each peripheral has a maximum of one in-edge since they can only derive from one
        # peripheral. Therefore, once they are encountered here they have no remaining dependencies.
        no_dep_peripherals.extend(dep_graph[peripheral.name])
        dep_graph.pop(peripheral.name, None)

    if dep_graph:
        raise ValueError(
            "Unable to determine order in which peripherals are derived. "
            "This is likely caused either by a cycle in the "
            "'derivedFrom' attributes, or a 'derivedFrom' attribute pointing to a "
            "nonexistent peripheral."
        )

    return sorted_peripherals


def remove_registers(
    peripheral_element: bindings.PeripheralElement,
    remove: Mapping[str, Sequence[str]],
) -> None:
    """
    Remove clusters/registers from a peripheral by deleting the nodes from the XML tree itself.

    :param peripheral_element: Peripheral node to filter registers from.
    :param remove:
    """
    registers = peripheral_element._registers
    if registers is None:
        # Skip if the node has no <registers> node (permitted on derived peripherals)
        return

    for pattern_str, paths in remove.items():
        if re.fullmatch(pattern_str, peripheral_element.name) is None:
            continue

        for path in paths:
            xpath = "." + "".join((f"/*[name='{p}']" for p in path.split(".")))
            nodes = typing.cast(Iterable[ET._Element], registers.xpath(xpath))
            for node in nodes:
                if (parent := node.getparent()) is not None:
                    parent.remove(node)


def svd_element_repr(
    klass: type,
    name: str,
    /,
    *,
    address: Optional[int] = None,
    length: Optional[int] = None,
    content: Optional[int] = None,
    content_max_width: int = 32,
    bool_props: Iterable[Any] = (),
    kv_props: Mapping[Any, Any] = MappingProxyType({}),
) -> str:
    """
    Common pretty print function for SVD elements.

    :param klass: Class of the element.
    :param name: Name of the element.
    :param address: Address of the element.
    :param length: Address of the element.
    :param content: Length of the element.
    :param content_max_width: Available width of the element, used to zero-pad the value.
    :param bool_props: Additional arguments to include in the pretty print.
    :param kv_props: Additional keyword arguments to include in the pretty print.

    :return: Pretty printed string representing the element.
    """

    address_str: str = f" @ 0x{address:08x}" if address is not None else ""
    length_str: str = f"<{length}>" if length is not None else ""
    value_str: str

    if content is not None:
        leading_zeros: str = "0" * ((content_max_width - content.bit_length()) // 4)
        value_str = f" = 0x{leading_zeros}{content:x}"
    else:
        value_str = ""

    if bool_props or kv_props:
        bool_props_str: str = (
            f"{', '.join(f'{v!s}' for v in bool_props)}" if bool_props else ""
        )
        kv_props_str: str = (
            f"{', '.join(f'{k}: {v!s}' for k, v in kv_props.items())})"
            if kv_props
            else ""
        )
        props_str = f" ({bool_props_str}{', ' if kv_props else ''}{kv_props_str})"
    else:
        props_str = ""

    return (
        f"[{name}{length_str}{address_str}{value_str}{props_str} {{{klass.__name__}}}]"
    )


def strip_suffix(word: str, suffix: str) -> str:
    """
    Remove the given suffix from the word, if present.

    :param word: String to strip prefixes and suffixes from.
    :param suffix: Suffix to strip.

    :return: word without the suffix.
    """

    if word.endswith(suffix):
        word = word[: -len(suffix)]

    return word


ItemT = TypeVar("ItemT")


class ChildIter(Reversible[ItemT]):
    """Helper class used as a generic reversible iterator"""

    def __init__(self, keys: Reversible, getter: Callable[[Any], ItemT]) -> None:
        self._keys = keys
        self._getter = getter

    def __iter__(self) -> Iterator[ItemT]:
        for k in self._keys:
            yield self._getter(k)

    def __reversed__(self) -> Iterator[ItemT]:
        for k in reversed(self._keys):
            yield self._getter(k)


class LazyFixedMapping(Mapping[str, ItemT]):
    """
    A mapping that lazily constructs its values.
    The set of keys is fixed at construction time - this ensures consistent ordering during
    iteration.
    """

    def __init__(
        self,
        keys: Iterable[str],
        factory: Callable[[str], ItemT],
        **kwargs: Any,
    ) -> None:
        """
        :param keys: Keys contained in the mapping.
        :param factory: Factory function which is called to initialize
                        new elements.
        """
        self._storage: Dict[str, Optional[ItemT]] = {key: None for key in keys}
        self._factory = factory

        super().__init__(**kwargs)

    def __getitem__(self, key: Union[str, Sequence[Union[str, Any]]], /) -> ItemT:
        remaining_key: Sequence[Union[str, Any]]

        if isinstance(key, str):
            this_key = key
            remaining_key = ()
        else:
            this_key = key[0]
            remaining_key = key[1:] if len(key) > 1 else ()

        if not isinstance(key, str):
            raise TypeError(f"Key {key} has invalid type {type(key)}. Expected a str.")

        value = self._storage[this_key]
        if value is None:
            value = self._factory(key)
            self._storage[this_key] = value

        if remaining_key:
            if not hasattr(value, "__getitem__"):
                raise LookupError(
                    f"̈́Invalid key for {value!r}: {remaining_key}. "
                    "The object is not subscriptable."
                )
            else:
                return value[remaining_key]
        else:
            return value

    def __iter__(self) -> Iterator[str]:
        return iter(self._storage)

    def __contains__(self, key: Any) -> bool:
        return key in self._storage

    def __len__(self) -> int:
        return len(self._storage)


class _End:
    ...


# Sentinel value for the end of an iterator
_END = _End()


def iter_merged(a: Iterable[ItemT], b: Iterable[ItemT], key: Callable[[ItemT], Any]) -> Iterator[ItemT]:
    """
    Iterator that merges two sorted iterables.
    It is assumed that both iterables are sorted - passing unsorted iterables will not
    cause an error.
    """

    iter_a = iter(a)
    iter_b = iter(b)

    item_a = next(iter_a, _END)
    item_b = next(iter_b, _END)

    while True:
        if not isinstance(item_a, _End) and not isinstance(item_b, _End):
            key_a = key(item_a)
            key_b = key(item_b)
            if key_a <= key_b:
                yield item_a
                item_a = next(iter_a, _END)
            else:
                yield item_b
                item_b = next(iter_b, _END)
        elif not isinstance(item_a, _End):
            yield item_a
            item_a = next(iter_a, _END)
        elif not isinstance(item_b, _End):
            yield item_b
            item_b = next(iter_b, _END)
        else:
            break



#
# Copyright (c) 2023 Nordic Semiconductor ASA
#
# SPDX-License-Identifier: Apache-2.0
#

"""
"Low-level" read-only Python representation of the SVD format that aims to represent the full SVD
document. Each type of element in the SVD XML tree is represented by a class in this module.
The class properties correspond more or less directly to the XML elements/attributes,
with some abstractions and simplifications added for convenience.

Based on CMSIS-SVD schema v1.3.9.
"""

from __future__ import annotations

import dataclasses as dc
import enum
import typing
from dataclasses import dataclass
from typing import Iterator, NamedTuple, Optional, Protocol, Sequence, Union

from lxml import objectify
from lxml.objectify import BoolElement, StringElement
from typing_extensions import TypeGuard, override

from ._bindings import (
    SELF_CLASS,
    Attr,
    BindingRegistry,
    CaseInsensitiveStrEnum,
    Elem,
    SvdElement,
    SvdIntElement,
    get_binding_elem_props,
    iter_element_children,
    make_enum_wrapper,
    to_bool,
    to_int,
)

# Container for classes that represent non-leaf elements in the SVD XML tree.
BINDING_REGISTRY = BindingRegistry()

# Alias for the BINDING_REGISTRY.add for convenience.
binding = BINDING_REGISTRY.add

# Alias for the BINDING_REGISTRY.bindings for convenience.
BINDINGS = BINDING_REGISTRY.bindings

# Utility function for the parse module
get_binding_elem_props = get_binding_elem_props


@enum.unique
class Access(CaseInsensitiveStrEnum):
    """
    Access rights for a given register or field.
    See "accessType" in the SVD schema.
    """

    # Read access is permitted. Write operations have an undefined result.
    READ_ONLY = "read-only"
    # Write access is permitted. Read operations have an undefined result.
    WRITE_ONLY = "write-only"
    # Read and write accesses are permitted.
    READ_WRITE = "read-write"
    # Only the first write after reset has an effect. Read operations have an undefined results.
    WRITE_ONCE = "writeOnce"
    # Only the first write after reset has an effect. Read access is permitted.
    READ_WRITE_ONCE = "read-writeOnce"


AccessElement = make_enum_wrapper(Access)


@enum.unique
class ReadAction(CaseInsensitiveStrEnum):
    """
    Side effect following a read operation of a given register or field.
    See "readActionType" in the SVD schema.
    """

    # The register/field is set to zero following a read operation.
    CLEAR = "clear"
    # The register/field is set to ones following a read operation.
    SET = "set"
    # The register/field is modified by a read operation.
    MODIFY = "modify"
    # A dependent resource is modified by a read operation.
    MODIFY_EXTERNAL = "modifyExternal"


ReadActionElement = make_enum_wrapper(ReadAction)


@enum.unique
class Endian(CaseInsensitiveStrEnum):
    """
    Processor endianness.
    See "endianType" in the SVD schema.
    """

    # Little endian
    LITTLE = "little"
    # Big endian
    BIG = "big"
    # Endianness is configurable for the device, taking effect on the next reset.
    SELECTABLE = "selectable"
    # Neither big nor little endian
    OTHER = "other"


EndianElement = make_enum_wrapper(Endian)


@enum.unique
class SauAccess(CaseInsensitiveStrEnum):
    """
    SAU region access type.
    See "sauAccessType" in the SVD schema.
    """

    # Non-secure accessible
    NON_SECURE = "n"
    # Secure callable
    SECURE_CALLABLE = "c"


SauAccessElement = make_enum_wrapper(SauAccess)


@enum.unique
class AddressBlockUsage(CaseInsensitiveStrEnum):
    """
    Defined usage type of a peripheral address block.
    See "addressBlockType" in the SVD schema.
    """

    REGISTER = "registers"
    BUFFER = "buffer"
    RESERVED = "reserved"


AddressBlockUsageElement = make_enum_wrapper(AddressBlockUsage)


@enum.unique
class Protection(CaseInsensitiveStrEnum):
    """
    Security privilege required to access an address region.
    See "protectionStringType" in the SVD schema.
    """

    # Secure permission required for access
    SECURE = "s"
    # Non-secure or secure permission required for access
    NON_SECURE = "n"
    # Privileged permission required for access
    PRIVILEGED = "p"


ProtectionElement = make_enum_wrapper(Protection)


@enum.unique
class EnumUsage(CaseInsensitiveStrEnum):
    """
    Usage of an enumerated value.
    See "enumUsageType" in the SVD schema.
    """

    # The value is relevant for read operations.
    READ = "read"
    # The value is relevant for write operations.
    WRITE = "write"
    # The value is relevant for read and write operations.
    READ_WRITE = "read-write"


EnumUsageElement = make_enum_wrapper(EnumUsage)


@enum.unique
class WriteAction(CaseInsensitiveStrEnum):
    """
    Side effect following a write operation of a given register or field.
    See "modifiedWriteValuesType" in the SVD schema.
    """

    # Bits written to one are set to zero in the register/field.
    ONE_TO_CLEAR = "oneToClear"
    # Bits written to one are set to one in the register/field.
    ONE_TO_SET = "oneToSet"
    # Bits written to one are inverted in the register/field.
    ONE_TO_TOGGLE = "oneToToggle"
    # Bits written to zero are set to zero in the register/field.
    ZERO_TO_CLEAR = "zeroToClear"
    # Bits written to zero are set to one in the register/field.
    ZERO_TO_SET = "zeroToSet"
    # Bits written to zero are inverted in the register/field.
    ZERO_TO_TOGGLE = "zeroToToggle"
    # All bits are set to zero on writing to the register/field.
    CLEAR = "clear"
    # All bits are set to one on writing to the register/field.
    SET = "set"
    # All bits are modified on writing to the register/field.
    MODIFY = "modify"


ModifiedWriteValuesElement = make_enum_wrapper(WriteAction)


@enum.unique
class DataType(CaseInsensitiveStrEnum):
    """
    Data types defined in the SVD specification.
    See "dataTypeType" in the SVD schema.
    """

    UINT8_T = "uint8_t"
    UINT16_T = "uint16_t"
    UINT32_T = "uint32_t"
    UINT64_T = "uint64_t"
    INT8_T = "int8_t"
    INT16_T = "int16_t"
    INT32_T = "int32_t"
    INT64_T = "int64_t"
    UINT8_PTR_T = "uint8_t *"
    UINT16_PTR_T = "uint16_t *"
    UINT32_PTR_T = "uint32_t *"
    UINT64_PTR_T = "uint64_t *"
    INT8_PTR_T = "int8_t *"
    INT16_PTR_T = "int16_t *"
    INT32_PTR_T = "int32_t *"
    INT64_PTR_T = "int64_t *"


DataTypeElement = make_enum_wrapper(DataType)


@enum.unique
class CpuName(CaseInsensitiveStrEnum):
    """
    CPU names defined in the SVD specification.
    See "cpuNameType" in the SVD schema.
    """

    CM0 = "CM0"
    CM0_PLUS_ = "CM0PLUS"
    CM0_PLUS = "CM0+"
    CM1 = "CM1"
    CM3 = "CM3"
    CM4 = "CM4"
    CM7 = "CM7"
    CM23 = "CM23"
    CM33 = "CM33"
    CM35P = "CM35P"
    CM55 = "CM55"
    CM85 = "CM85"
    SC000 = "SC000"
    SC300 = "SC300"
    ARMV8MML = "ARMV8MML"
    ARMV8MBL = "ARMV8MBL"
    ARMV81MML = "ARMV81MML"
    CA5 = "CA5"
    CA7 = "CA7"
    CA8 = "CA8"
    CA9 = "CA9"
    CA15 = "CA15"
    CA17 = "CA17"
    CA53 = "CA53"
    CA57 = "CA57"
    CA72 = "CA72"
    SMC1 = "SMC1"
    OTHER = "other"


CpuNameElement = make_enum_wrapper(CpuName)


@binding
class RangeWriteConstraint(SvdElement):
    """Value range constraint for a register or field."""

    TAG: str = "range"

    # Minimum permitted value
    minimum: Elem[int] = Elem("minimum", SvdIntElement)

    # Maximum permitted value
    maximum: Elem[int] = Elem("maximum", SvdIntElement)


@enum.unique
class WriteConstraint(enum.Enum):
    """Type of write constraint for a register or field."""

    # Only the last read value can be written.
    WRITE_AS_READ = enum.auto()
    # Only enumerated values can be written.
    USE_ENUMERATED_VALUES = enum.auto()
    # Only values within a given range can be written.
    RANGE = enum.auto()


@binding
class WriteConstraintElement(SvdElement):
    """Constraint on permitted values in a register or field."""

    TAG: str = "writeConstraint"

    # Value range constraint
    value_range: Elem[Optional[RangeWriteConstraint]] = Elem(
        "range", RangeWriteConstraint, default=None
    )

    def as_enum(self) -> Optional[WriteConstraint]:
        """Return the write constraint as an enum value."""
        if self._write_as_read:
            return WriteConstraint.WRITE_AS_READ
        if self._use_enumerated_values:
            return WriteConstraint.USE_ENUMERATED_VALUES
        if self.value_range is not None:
            return WriteConstraint.RANGE
        return None

    # (internal) If true, only the last read value can be written.
    _write_as_read: Elem[bool] = Elem("writeAsRead", BoolElement, default=False)

    # (internal) If true, only enumerated values can be written.
    _use_enumerated_values: Elem[bool] = Elem(
        "useEnumeratedValues", BoolElement, default=False
    )


@binding
class SauRegion(SvdElement):
    """Predefined Secure Attribution Unit (SAU) region."""

    TAG: str = "region"

    # If true, the SAU region is enabled.
    enabled: Attr[bool] = Attr("enabled", converter=to_bool, default=True)

    # Name of the SAU region.
    name: Attr[Optional[str]] = Attr("name", default=None)

    # Base address of the SAU region.
    base_address: Elem[int] = Elem("base", SvdIntElement)

    # Limit address of the SAU region.
    limit: Elem[int] = Elem("limit", SvdIntElement)

    # Access permissions of the SAU region.
    access: Elem[SauAccess] = Elem("access", SauAccessElement)


@binding
class SauRegionsConfig(SvdElement):
    """Container for predefined Secure Attribution Unit (SAU) regions."""

    TAG: str = "sauRegions"

    # If true, the SAU is enabled.
    enabled: Attr[bool] = Attr("enabled", converter=to_bool, default=True)

    # Default protection for disabled SAU regions.
    protection_when_disabled: Attr[Optional[Protection]] = Attr(
        "name", converter=Protection, default=None
    )

    @property
    def regions(self) -> Iterator[SauRegion]:
        """Iterate over all predefined SAU regions."""
        it = iter_element_children(self, SauRegion.TAG)
        return typing.cast(Iterator[SauRegion], it)

    # (internal) SAU regions.
    _region: Elem[SauRegion] = Elem("region", SauRegion)


@binding
class Cpu(SvdElement):
    """Description of the device processor."""

    TAG: str = "cpu"

    # CPU name. See CpuName for possible values.
    name: Elem[CpuName] = Elem("name", CpuNameElement)

    # CPU hardware revision with the format "rNpM".
    revision: Elem[str] = Elem("revision", StringElement)

    # Default endianness of the CPU.
    endian: Elem[Endian] = Elem("endian", EndianElement)

    # True if the CPU has a memory protection unit (MPU).
    has_mpu: Elem[Optional[bool]] = Elem("mpuPresent", BoolElement, default=None)

    # True if the CPU has a floating point unit (FPU).
    has_fpu: Elem[Optional[bool]] = Elem("fpuPresent", BoolElement, default=None)

    # True if the CPU has a double precision floating point unit (FPU).
    fpu_is_double_precision: Elem[Optional[bool]] = Elem(
        "fpuDP", BoolElement, default=None
    )

    # True if the CPU implements the SIMD DSP extensions.
    has_dsp: Elem[Optional[bool]] = Elem("dspPresent", BoolElement, default=None)

    # True if the CPU has an instruction cache.
    has_icache: Elem[Optional[bool]] = Elem("icachePresent", BoolElement, default=None)

    # True if the CPU has a data cache.
    has_dcache: Elem[Optional[bool]] = Elem("dcachePresent", BoolElement, default=None)

    # True if the CPU has an instruction tightly coupled memory (ITCM).
    has_ictm: Elem[Optional[bool]] = Elem("itcmPresent", BoolElement, default=None)

    # True if the CPU has a data tightly coupled memory (DTCM).
    has_dctm: Elem[Optional[bool]] = Elem("dtcmPresent", BoolElement, default=None)

    # True if the CPU has a Vector Table Offset Register (VTOR).
    has_vtor: Elem[bool] = Elem("vtorPresent", BoolElement, default=True)

    # Bit width of interrupt priority levels in the Nested Vectored Interrupt Controller (NVIC).
    num_nvic_priority_bits: Elem[int] = Elem("nvicPrioBits", SvdIntElement)

    # True if the CPU has a vendor-specific SysTick Timer.
    # If False, the Arm-defined System Tick Timer is used.
    has_vendor_systick: Elem[bool] = Elem("vendorSystickConfig", BoolElement)

    # Maximum interrupt number in the CPU plus one.
    num_interrupts: Elem[Optional[int]] = Elem(
        "deviceNumInterrupts", SvdIntElement, default=None
    )

    # Number of supported Secure Attribution Unit (SAU) regions.
    num_sau_regions: Elem[int] = Elem("sauNumRegions", SvdIntElement, default=0)

    # Predefined Secure Attribution Unit (SAU) regions, if any.
    preset_sau_regions: Elem[Optional[SauRegionsConfig]] = Elem(
        "sauRegionsConfig", SauRegionsConfig, default=None
    )


@binding
class AddressBlock(SvdElement):
    """Address range mapped to a peripheral."""

    TAG: str = "addressBlock"

    # Start address of the address block, relative to the peripheral base address.
    offset: Elem[int] = Elem("offset", SvdIntElement)

    # Number of address unit bits covered by the address block.
    size: Elem[int] = Elem("size", SvdIntElement)

    # Address block usage. See AddressBlockUsage for possible values.
    usage: Elem[AddressBlockUsage] = Elem("usage", AddressBlockUsageElement)

    # Protection level for the address block.
    protection: Elem[Optional[Protection]] = Elem("protection", ProtectionElement)


class DerivedMixin(objectify.ObjectifiedElement):
    """Common functionality for elements that contain a SVD 'derivedFrom' attribute."""

    # Name of the element that this element is derived from.
    derived_from: Attr[Optional[str]] = Attr("derivedFrom", default=None)

    @property
    def is_derived(self) -> bool:
        """Return True if the element is derived from another element."""
        return self.derived_from is not None


@binding
class EnumeratedValue(SvdElement):
    """Value definition for a field."""

    TAG: str = "enumeratedValue"

    # Name of the enumerated value.
    name: Elem[str] = Elem("name", StringElement)

    # Description of the enumerated value.
    description: Elem[Optional[str]] = Elem("description", StringElement, default=None)

    # Value of the enumerated value.
    value: Elem[int] = Elem("value", SvdIntElement)

    # True if the enumerated value is the default value of the field.
    is_default: Elem[bool] = Elem("isDefault", BoolElement, default=False)


@binding
class Enumeration(SvdElement, DerivedMixin):
    """Container for enumerated values."""

    TAG: str = "enumeratedValues"

    # Name of the enumeration.
    name: Elem[Optional[str]] = Elem("name", StringElement, default=None)

    # Identifier of the enumeration in the device header file.
    header_enum_name: Elem[Optional[str]] = Elem(
        "headerEnumName", StringElement, default=None
    )

    # Description of which types of operations the enumeration is used for.
    usage: Elem[EnumUsage] = Elem(
        "usage", EnumUsageElement, default=EnumUsage.READ_WRITE
    )

    @property
    def enums(self) -> Iterator[EnumeratedValue]:
        """Iterate over all enumerated values."""
        it = iter_element_children(self, EnumeratedValue.TAG)
        return typing.cast(Iterator[EnumeratedValue], it)

    # (internal) Enumerated values
    _enumerated_values: Elem[EnumeratedValue] = Elem("enumeratedValue", EnumeratedValue)


@binding
class DimArrayIndex(SvdElement):
    """Description of the index used for an array of registers."""

    TAG: str = "dimArrayIndex"

    # The base name of enumerations.
    header_enum_name: Elem[Optional[str]] = Elem(
        "headerEnumName", StringElement, default=None
    )

    # Values contained in the enumeration.
    enumerated_values: Elem[Iterator[EnumeratedValue]] = Elem(
        "enumeratedValue", EnumeratedValue
    )


@binding
class Interrupt(SvdElement):
    """Peripheral interrupt description."""

    TAG: str = "interrupt"

    # Name of the interrupt.
    name: Elem[str] = Elem("name", StringElement)

    # Description of the interrupt.
    description: Elem[Optional[str]] = Elem("description", StringElement, default=None)

    # Interrupt number.
    value: Elem[int] = Elem("value", SvdIntElement)


class BitRange(NamedTuple):
    """Bit range of a field."""

    # Bit offset of the field.
    offset: int

    # Bit width of the field.
    width: int


@binding
class FieldElement(SvdElement, DerivedMixin):
    """SVD field element."""

    TAG: str = "field"

    # Name of the field.
    name: Elem[str] = Elem("name", StringElement)

    # Description of the field.
    description: Elem[Optional[str]] = Elem("description", StringElement, default=None)

    # Access rights of the field.
    access: Elem[Optional[Access]] = Elem("access", AccessElement, default=None)

    # Side effect when writing to the field.
    modified_write_values: Elem[Optional[WriteAction]] = Elem(
        "modifiedWriteValue", ModifiedWriteValuesElement, default=None
    )

    # Constraints on writing to the field.
    write_constraint: Elem[Optional[WriteConstraintElement]] = Elem(
        "writeConstraint", WriteConstraintElement, default=None
    )

    # Side effect when reading from the field.
    read_action: Elem[Optional[ReadAction]] = Elem(
        "readAction", ReadActionElement, default=None
    )

    # Permitted values of the field.
    enumerated_values: Elem[Optional[Enumeration]] = Elem(
        "enumeratedValues", Enumeration, default=None
    )

    @property
    def bit_range(self) -> BitRange:
        """
        Bit range of the field.
        :return: Tuple of the field's bit offset and bit width.
        """

        if self._lsb is not None and self._msb is not None:
            return BitRange(offset=self._lsb, width=self._msb - self._lsb + 1)

        if self._bit_offset is not None:
            width = self._bit_width if self._bit_width is not None else 32
            return BitRange(offset=self._bit_offset, width=width)

        if self._bit_range is not None:
            msb_string, lsb_string = self._bit_range[1:-1].split(":")
            msb, lsb = to_int(msb_string), to_int(lsb_string)
            return BitRange(offset=lsb, width=msb - lsb + 1)

        return BitRange(offset=0, width=32)

    # (internal) Least significant bit of the field, if specified in the bitRangeLsbMsbStyle style.
    _lsb: Elem[Optional[int]] = Elem("lsb", SvdIntElement, default=None)

    # (internal) Most significant bit of the field, if specified in the bitRangeLsbMsbStyle style.
    _msb: Elem[Optional[int]] = Elem("msb", SvdIntElement, default=None)

    # (internal) Bit offset of the field, if specified in the bitRangeOffsetWidthStyle style.
    _bit_offset: Elem[Optional[int]] = Elem("bitOffset", SvdIntElement, default=None)

    # (internal) Bit width of the field, if specified in the bitRangeOffsetWidthStyle style.
    _bit_width: Elem[Optional[int]] = Elem("bitWidth", SvdIntElement, default=None)

    # (internal) Bit range of the field, given in the form "[msb:lsb]", if specified in the
    # bitRangePattern style.
    _bit_range: Elem[Optional[str]] = Elem("bitRange", StringElement, default=None)

    def __repr__(self) -> str:
        try:
            name = self.name
        except Exception:
            name = None

        return super()._repr(props={"name": name})


@binding
class FieldsElement(SvdElement):
    """Container for SVD field elements."""

    TAG: str = "fields"

    # Field elements.
    field: Elem[FieldElement] = Elem("field", FieldElement)


@dataclass
class RegisterProperties:
    """Common SVD device/peripheral/register level properties."""

    # Size of the register in bits.
    size: Optional[int]

    # Access rights of the register.
    access: Optional[Access]

    # Protection level of the register.
    protection: Optional[Protection]

    # Reset value of the register.
    reset_value: Optional[int]

    # Reset mask of the register.
    reset_mask: Optional[int]


class FullRegisterProperties(Protocol):
    """Protocol that describes a fully defined register properties object."""

    size: int
    access: Access
    protection: Optional[Protection]
    reset_value: int
    reset_mask: int

    @staticmethod
    def is_full(props: RegisterProperties) -> TypeGuard[FullRegisterProperties]:
        """Check if the given register properties has all the required fields set."""
        return all(
            f is not None
            for f in (
                props.size,
                props.access,
                props.reset_value,
                props.reset_mask,
            )
        )


class RegisterPropertiesGroupMixin(objectify.ObjectifiedElement):
    """Common functionality for elements that contain a SVD 'registerPropertiesGroup'."""

    @property
    def register_properties(self) -> RegisterProperties:
        """Register properties specified in the element itself."""
        return self.get_register_properties()

    def get_register_properties(
        self, base_props: Optional[RegisterProperties] = None
    ) -> RegisterProperties:
        """
        Get the register properties of the element, optionally inheriting from a
        base set of properties.
        """
        if base_props is None:
            return RegisterProperties(
                size=self._size,
                access=self._access,
                protection=self._protection,
                reset_value=self._reset_value,
                reset_mask=self._reset_mask,
            )

        return RegisterProperties(
            size=self._size if self._size is not None else base_props.size,
            access=self._access if self._access is not None else base_props.access,
            protection=(
                self._protection
                if self._protection is not None
                else base_props.protection
            ),
            reset_value=(
                self._reset_value
                if self._reset_value is not None
                else base_props.reset_value
            ),
            reset_mask=(
                self._reset_mask
                if self._reset_mask is not None
                else base_props.reset_mask
            ),
        )

    _size: Elem[Optional[int]] = Elem("size", SvdIntElement, default=None)
    _access: Elem[Optional[Access]] = Elem("access", AccessElement, default=None)
    _protection: Elem[Optional[Protection]] = Elem(
        "protection", ProtectionElement, default=None
    )
    _reset_value: Elem[Optional[int]] = Elem("resetValue", SvdIntElement, default=None)
    _reset_mask: Elem[Optional[int]] = Elem("resetMask", SvdIntElement, default=None)


@dataclass
class Dimensions:
    """Dimensions of a repeated SVD element."""

    # Number of times the element is repeated.
    length: int

    # Increment between each element.
    step: int

    def to_range(self) -> Sequence[int]:
        """Convert to a range of offsets"""
        return range(0, (self.length - 1) * self.step + 1, self.step)


class DimElementGroupMixin(objectify.ObjectifiedElement):
    """Common functionality for elements that contain a SVD 'dimElementGroup'."""

    # Index of the element, if it is repeated.
    dim_index: Elem[Optional[int]] = Elem("dimIndex", SvdIntElement, default=None)

    # Name of the dimension, if it is repeated.
    dim_name: Elem[Optional[str]] = Elem("dimName", StringElement, default=None)

    # Array index of the element, if it is repeated.
    dim_array_index: Elem[Optional[DimArrayIndex]] = Elem(
        "dimArrayIndex", DimArrayIndex, default=None
    )

    @property
    def dimensions(self) -> Optional[Dimensions]:
        """Get the dimensions of the element, if it is repeated."""
        if self._dim is None or self._dim_increment is None:
            return None

        return Dimensions(
            length=self._dim,
            step=self._dim_increment,
        )

    _dim: Elem[Optional[int]] = Elem("dim", SvdIntElement, default=None)
    _dim_increment: Elem[Optional[int]] = Elem(
        "dimIncrement", SvdIntElement, default=None
    )


@binding
class RegisterElement(
    SvdElement,
    DimElementGroupMixin,
    RegisterPropertiesGroupMixin,
    DerivedMixin,
):
    """SVD register element."""

    TAG: str = "register"

    # Name of the register.
    name: Elem[str] = Elem("name", StringElement)

    # Display name of the register.
    display_name: Elem[Optional[str]] = Elem("displayName", StringElement, default=None)

    # Description of the register.
    description: Elem[Optional[str]] = Elem("description", StringElement, default=None)

    # Alternate group of the register.
    alternate_group: Elem[Optional[str]] = Elem(
        "alternateGroup", StringElement, default=None
    )

    # Name of a different register that corresponds to this register.
    alternate_register: Elem[Optional[str]] = Elem(
        "alternateRegister", StringElement, default=None
    )

    # Address offset of the register, relative to the parent element.
    offset: Elem[Optional[int]] = Elem("addressOffset", SvdIntElement, default=None)

    # C data type to use when accessing the register.
    data_type: Elem[Optional[DataType]] = Elem(
        "dataType", DataTypeElement, default=None
    )

    # Side effect of writing the register.
    modified_write_values: Elem[WriteAction] = Elem(
        "modifiedWriteValues", ModifiedWriteValuesElement, default=WriteAction.MODIFY
    )

    # Write constraint of the register.
    write_constraint: Elem[Optional[WriteConstraintElement]] = Elem(
        "writeConstraint", WriteConstraintElement, default=None
    )

    # Side effect of reading the register.
    read_action: Elem[Optional[ReadAction]] = Elem(
        "readAction", ReadActionElement, default=None
    )

    @property
    def fields(self) -> Iterator[FieldElement]:
        """Iterator over the fields of the register."""
        it = iter_element_children(self._fields, FieldElement.TAG)
        return typing.cast(Iterator[FieldElement], it)

    # (internal) Fields of the register.
    _fields: Elem[Optional[FieldElement]] = Elem("fields", FieldsElement, default=None)

    def __repr__(self) -> str:
        try:
            name = self.name
        except Exception:
            name = None

        return super()._repr(props={"name": name})


@binding
class ClusterElement(
    SvdElement,
    DimElementGroupMixin,
    RegisterPropertiesGroupMixin,
    DerivedMixin,
):
    """SVD cluster element."""

    TAG: str = "cluster"

    # Name of the cluster.
    name: Elem[str] = Elem("name", StringElement)

    # Description of the cluster.
    description: Elem[Optional[str]] = Elem("description", StringElement, default=None)

    # Name of a different cluster that corresponds to this cluster.
    alternate_cluster: Elem[Optional[str]] = Elem(
        "alternateCluster", StringElement, default=None
    )

    # Name of the C struct used to represent the cluster.
    header_struct_name: Elem[Optional[str]] = Elem(
        "headerStructName", StringElement, default=None
    )

    # Address offset of the cluster, relative to the parent peripheral element.
    offset: Elem[Optional[int]] = Elem("addressOffset", SvdIntElement, default=None)

    @property
    def registers(self) -> Iterator[Union[RegisterElement, ClusterElement]]:
        """Iterator over the registers and clusters that are direct children of this cluster."""
        it = iter_element_children(self, RegisterElement.TAG, ClusterElement.TAG)
        return typing.cast(Iterator[Union[RegisterElement, ClusterElement]], it)

    # (internal) Register elements in the cluster.
    _register: Elem[Optional[RegisterElement]] = Elem(
        "register", RegisterElement, default=None
    )

    # (internal) Cluster elements in the cluster.
    _cluster: Elem[Optional[ClusterElement]] = Elem("cluster", SELF_CLASS, default=None)

    def __repr__(self) -> str:
        try:
            name = self.name
        except Exception:
            name = None

        return super()._repr(props={"name": name})


@binding
class RegistersElement(SvdElement):
    """Container for SVD register/cluster elements."""

    TAG: str = "registers"

    # Cluster elements in the container.
    cluster: Elem[Optional[ClusterElement]] = Elem(
        "cluster", ClusterElement, default=None
    )

    # Register elements in the container.
    register: Elem[Optional[RegisterElement]] = Elem(
        "register", RegisterElement, default=None
    )


@binding
class PeripheralElement(
    SvdElement,
    DimElementGroupMixin,
    RegisterPropertiesGroupMixin,
    DerivedMixin,
):
    """SVD peripheral element."""

    TAG: str = "peripheral"

    # Name of the peripheral.
    name: Elem[str] = Elem("name", StringElement)

    # Version of the peripheral.
    version: Elem[Optional[str]] = Elem("version", StringElement, default=None)

    # Description of the peripheral.
    description: Elem[Optional[str]] = Elem("description", StringElement, default=None)

    # Base address of the peripheral.
    base_address: Elem[int] = Elem("baseAddress", SvdIntElement)

    @property
    def interrupts(self) -> Iterator[Interrupt]:
        """Iterator over the interrupts of the peripheral."""
        it = iter_element_children(self, Interrupt.TAG)
        return typing.cast(Iterator[Interrupt], it)

    @property
    def address_blocks(self) -> Iterator[AddressBlock]:
        """Iterator over the address blocks of the peripheral."""
        it = iter_element_children(self, AddressBlock.TAG)
        return typing.cast(Iterator[AddressBlock], it)

    @property
    def registers(self) -> Iterator[Union[RegisterElement, ClusterElement]]:
        """Iterator over the registers and clusters that are direct children of this peripheral."""
        it =iter_element_children(
            self._registers, RegisterElement.TAG, ClusterElement.TAG
        )
        return typing.cast(Iterator[Union[RegisterElement, ClusterElement]], it)

    # Name of a different peripheral that corresponds to this peripheral.
    alternate_peripheral: Elem[Optional[str]] = Elem(
        "alternatePeripheral", StringElement, default=None
    )

    # Name of the group that the peripheral belongs to.
    group_name: Elem[Optional[str]] = Elem("groupName", StringElement, default=None)

    # String to prepend to the names of registers contained in the peripheral.
    prepend_to_name: Elem[Optional[str]] = Elem(
        "prependToName", StringElement, default=None
    )

    # String to append to the names of registers contained in the peripheral.
    append_to_name: Elem[Optional[str]] = Elem(
        "appendToName", StringElement, default=None
    )

    # Name of the C struct that represents the peripheral.
    header_struct_name: Elem[Optional[str]] = Elem(
        "headerStructName", StringElement, default=None
    )

    disable_condition: Elem[Optional[str]] = Elem(
        "disableCondition", StringElement, default=None
    )

    # (internal) Interrupt elements in the peripheral.
    _interrupts: Elem[Optional[Interrupt]] = Elem("interrupt", Interrupt, default=None)

    # (internal) Address block elements in the peripheral.
    _address_blocks: Elem[Optional[AddressBlock]] = Elem(
        "addressBlock", AddressBlock, default=None
    )

    # (internal) Register/cluster container.
    _registers: Elem[Optional[RegistersElement]] = Elem(
        "registers", RegistersElement, default=None
    )

    def __repr__(self) -> str:
        try:
            name = self.name
        except Exception:
            name = None

        props = {"name": name}

        if (derived_from := self.derived_from) is not None:
            props["derived_from"] = derived_from

        return super()._repr(props=props)


@binding
class PeripheralsElement(SvdElement):
    """Container for SVD peripheral elements."""

    TAG: str = "peripherals"

    # Peripheral elements in the container.
    peripheral: Elem[Optional[PeripheralElement]] = Elem(
        "peripheral", PeripheralElement, default=None
    )


@binding
class DeviceElement(SvdElement, RegisterPropertiesGroupMixin):
    """SVD device element."""

    TAG: str = "device"

    # Version of the CMSIS schema that the SVD file conforms to.
    schema_version: Attr[float] = Attr("schemaVersion", converter=float)

    # Name of the device.
    name: Elem[str] = Elem("name", StringElement)

    # Device series name.
    series: Elem[Optional[str]] = Elem("series", StringElement, default=None)

    # Version of the device.
    version: Elem[str] = Elem("version", StringElement)

    # Full device vendor name.
    vendor: Elem[Optional[str]] = Elem("vendor", StringElement, default=None)

    # Abbreviated device vendor name.
    vendor_id: Elem[Optional[str]] = Elem("vendorID", StringElement, default=None)

    # Description of the device.
    description: Elem[str] = Elem("description", StringElement)

    # The license to use for the device header file.
    license_text: Elem[Optional[str]] = Elem("licenseText", StringElement)

    # Description of the device processor.
    cpu: Elem[Optional[Cpu]] = Elem("cpu", Cpu, default=None)

    # Device header filename without extension.
    header_system_filename: Elem[Optional[str]] = Elem(
        "headerSystemFilename", StringElement, default=None
    )

    # String to prepend to all type definitions in the device header file.
    header_definitions_prefix: Elem[Optional[str]] = Elem(
        "headerDefinitionsPrefix", StringElement, default=None
    )

    # Number of data bits selected by each address.
    address_unit_bits: Elem[int] = Elem("addressUnitBits", SvdIntElement)

    # Width of the maximum data transfer supported by the device.
    width: Elem[int] = Elem("width", SvdIntElement)

    @property
    def peripherals(self) -> Iterator[PeripheralElement]:
        """Iterate over all peripherals in the device"""
        it = iter_element_children(self._peripherals, PeripheralElement.TAG)
        return typing.cast(Iterator[PeripheralElement], it)

    # (internal) Peripheral elements in the device.
    _peripherals: Elem[PeripheralElement] = Elem(
        "peripherals",
        PeripheralsElement,
    )



#
# Copyright (c) 2023 Nordic Semiconductor ASA
#
# SPDX-License-Identifier: Apache-2.0
#

"""
Internal basic profiling functionality.
"""

from __future__ import annotations

import inspect
import os
from time import perf_counter_ns
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Generic,
    List,
    Literal,
    NoReturn,
    Optional,
    Type,
    TypeVar,
    Union,
    overload,
)

from typing_extensions import Concatenate, ParamSpec, Self

PROFILING_ENABLED = bool(os.getenv("PROFILING", default=False))

T = TypeVar("T")
TimingReport = Dict[str, Union[None, float, List[float]]]

if TYPE_CHECKING or PROFILING_ENABLED:

    def create_object_report(obj: object) -> TimingReport:
        """Get a dictionary containing execution times for each timed method in the given object."""
        report = {}
        times: Union[None, float, List[float]]

        for method, timer_instance in obj.__dict__.get("_timedmethods", {}).items():
            if timer_instance.max_times <= 1:
                if timer_instance.times:
                    times = _ns_to_ms(timer_instance.times[0])
                else:
                    times = None
            else:
                times = [_ns_to_ms(t) for t in timer_instance.times]

            report[method.__name__] = times

        return report

    def timed_method(
        max_times: int = 10,
    ) -> Callable[[Callable[P, T]], Callable[P, T]]:
        """
        Decorator that records the execution times of a method.
        The execution times are stored in a circular buffer storing up to max times entries.

        :param max_times: Number of most recent times to keep.
        :return: TimedMethod object that wraps the method and records the execution
        time each time it is called.
        """

        def inner(method: Callable[P, T]) -> Callable[P, T]:
            return TimedMethod(method, max_times)  # type: ignore

        return inner

    def _ns_to_ms(time_ns: int) -> float:
        """Convert nanoseconds to miliseconds."""
        return time_ns / 1_000_000

    P = ParamSpec("P")
    O = TypeVar("O")

    class TimedMethod(Generic[O, P, T]):
        """
        Object wrapper around a method that records the most recent execution times
        of that method.
        """

        def __init__(
            self, method: Callable[Concatenate[object, P], T], max_times: int
        ) -> None:
            self._method: Callable[Concatenate[object, P], T] = method
            self._max_times: int = max_times

        def __call__(self, obj: O) -> T:
            """
            Call method that enables e.g. functools.cached_property to call the method
            through this object directly.
            """
            return self._get_instance(obj)()

        @overload
        def __get__(self, obj: Literal[None], owner: Optional[Type] = None) -> Self:
            ...

        @overload
        def __get__(
            self, obj: O, owner: Optional[Type] = None
        ) -> TimedMethodInstance[O, P, T]:
            ...

        def __get__(
            self, obj: Optional[O], owner: Any = None
        ) -> Union[TimedMethodInstance[O, P, T], Self]:
            """Get the child instance wrapper."""
            if obj is None:
                return self
            return self._get_instance(obj)

        def _get_instance(self, obj: O) -> TimedMethodInstance[O, P, T]:
            """Get the child wrapper instance, constructing it if accessed the first time."""
            obj_timed_methods = obj.__dict__.setdefault("_timedmethods", {})

            try:
                return obj_timed_methods[self._method]
            except KeyError:
                # First call; initialize the bound method
                instance = TimedMethodInstance(obj, self._method, self._max_times)
                obj_timed_methods[self._method] = instance
                return instance

    class TimedMethodInstance(Generic[O, P, T]):
        """
        Object wrapper around a method that records the most recent execution times
        of that method.
        """

        def __init__(
            self,
            obj: O,
            method: Callable[Concatenate[O, P], T],
            max_times: int,
        ) -> None:
            self._obj: O = obj
            self._method: Callable[Concatenate[O, P], T] = method
            self._max_times: int = max_times
            self._times: List[int] = []
            # Zero is the initial circular buffer head when the time buffer reaches its max capacity
            self._head: int = 0

        @property
        def times(self) -> List[int]:
            """
            A list of the most recent execution times, ordered from oldest to newest,
            with a size bounded by max_times.
            """
            if len(self._times) >= self._max_times:
                return self._times[self._head :] + self._times[: self._head]
            else:
                return self._times

        @property
        def max_times(self) -> int:
            """Maximum number of most recent execution times recorded."""
            return self._max_times

        def __call__(self, *args: P.args, **kwargs: P.kwargs) -> T:
            """Call the wrapped method and record the execution time."""
            t_start = perf_counter_ns()

            result = self._method(self._obj, *args, **kwargs)

            t_end = perf_counter_ns()
            self._add_time(t_end - t_start)

            return result

        def _add_time(self, time_ns: int) -> None:
            """
            Add a new time to the time buffer.
            The time buffer acts as a regular list before it reaches its capacity,
            and is used a circular buffer after that.
            """
            if len(self._times) >= self._max_times:
                self._times[self._head] = time_ns
                self._head = (self._head + 1) % self._max_times
            else:
                self._times.append(time_ns)

else:

    def create_object_report(obj: Any) -> NoReturn:
        raise RuntimeError(
            "Profiling info is not available as profiling is not enabled."
            "Enable profiling by setting the environment variable 'PROFILING'=1."
        )

    def timed_method(*args: Any, **kwargs: Any) -> Callable[[T], T]:
        # Just return the method back again
        return lambda x: x



#
# Copyright (c) 2023 Nordic Semiconductor ASA
#
# SPDX-License-Identifier: Apache-2.0
#

"""
Classes for refererencing SVD elements based on name.
"""

from __future__ import annotations

import re
from abc import ABC, abstractmethod
from functools import singledispatch
from itertools import chain
from typing import (
    Any,
    Generic,
    Iterable,
    List,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
    Union,
    overload,
)
from typing_extensions import Self


# Type of path part items 
PartT = TypeVar("PartT")


class AbstractEPath(ABC, Sequence[PartT]):
    """
    A type used to describe paths of named SVD elements.
    """

    @abstractmethod
    def __init__(self, *parts: Union[PartT, Sequence[PartT]]) -> None:
        """
        :param parts: Path segments
        """
        ...

    @property
    @abstractmethod
    def parts(self) -> Tuple[PartT, ...]:
        """:return: Path components."""
        ...

    @property
    @abstractmethod
    def name(self) -> Optional[str]:
        """:return: Name of the element pointed to by the path, including any array indices."""
        ...

    @property
    @abstractmethod
    def stem(self) -> Optional[str]:
        """:return: Name of the element pointed to by the path, excluding any array indices"""
        ...

    @property
    @abstractmethod
    def parent(self) -> Optional[AbstractEPath]:
        """:return: Path to the parent element of this path, if it exists."""
        ...

    def join(self, *other: Union[PartT, Sequence[PartT]]) -> Self:
        """:return: The path resulting from appending other to the end of this path. """
        return self.__class__(*self.parts, *other)

    @overload
    def __getitem__(self, item: int, /) -> PartT:
        ...

    @overload
    def __getitem__(self, item: slice, /) -> Self:
        ...

    def __getitem__(self, item: Union[int, slice], /) -> Union[PartT, Self]:
        if isinstance(item, slice):
            return self.__class__(*self.parts[item])
        else:
            return self.parts[item]

    def __len__(self) -> int:
        return len(self.parts)

    def __hash__(self) -> int:
        return hash(self.parts)

    def __eq__(self, other: Any) -> bool:
        return self.parts == other

    @abstractmethod
    def __repr__(self) -> str:
        ...


class FEPath(AbstractEPath[str]):
    """
    Path to a flat SVD register level element.
    An FEPath like "REGISTER_A.REGISTER_B" refers to the element with name "REGISTER_B" with a
    parent with the name "REGISTER_A".

    FEPaths refer to SVD register level elements by name without considering array dimensions,
    corresponding directly to how the elements are structured in the SVD file.
    """

    __slots__ = "_parts"

    def __init__(self, *parts: Union[str, Sequence[str]]) -> None:
        split_parts: List[str] = []

        for part in parts:
            if isinstance(part, str):
                split_parts.extend(part.split("."))
            elif isinstance(part, FEPath):
                split_parts.extend(part)
            else:
                sub_parts = (p.split(".") for p in part)
                split_parts.extend(chain.from_iterable(sub_parts))

        if not split_parts:
            raise ValueError(f"Empty {self.__class__.__name__} not allowed")

        if any(not p for p in split_parts):
            raise ValueError(f"Invalid {self.__class__.__name__} parts: {parts}")

        self._parts: Tuple[str, ...] = tuple(split_parts)

    @property
    def parts(self) -> Tuple[str, ...]:
        return self._parts

    @property
    def name(self) -> str:
        return self.parts[-1]

    @property
    def parent(self) -> Optional[FEPath]:
        if len(self.parts) <= 1:
            return None
        return FEPath(*self.parts[:-1])

    @property
    def stem(self) -> str:
        return self.name

    def to_xpath(self) -> str:
        """:return: An XPath expression that can be used to locate XML elements having this path"""
        return "." + "".join((f"/*[name='{p}']" for p in self.parts))

    def __repr__(self) -> str:
        return ".".join(self.parts)


class EPath(AbstractEPath[Union[str, int]]):
    """
    Path to a SVD element.
    """

    __slots__ = "_parts"

    def __init__(self, *parts: Union[str, int, Sequence[Union[str, int]]]) -> None:
        processed_parts = self._process_parts(parts)

        if not processed_parts:
            raise ValueError(f"Empty {self.__class__.__name__} not allowed")

        self._parts: Tuple[Union[str, int], ...] = tuple(processed_parts)

    @property
    def parts(self) -> Tuple[Union[str, int], ...]:
        return self._parts

    @property
    def name(self) -> str:
        for i in reversed(range(len(self._parts))):
            if isinstance(self._parts[i], str):
                return self._format_parts(self._parts[i:])
        assert False

    @property
    def stem(self) -> str:
        for part in reversed(self._parts):
            if isinstance(part, str):
                return part
        assert False

    @property
    def parent(self) -> Optional[EPath]:
        if len(self._parts) <= 1:
            return None
        return EPath(*self._parts[:-1])

    @property
    def element_index(self) -> Optional[int]:
        """Index of the register in the parent array, if applicable."""
        if not isinstance(self[-1], int):
            return None
        return self[-1]

    def to_flat(self) -> FEPath:
        """Convert the regular path to the equivalent flat path."""
        return FEPath(*(p for p in self.parts if not isinstance(p, int)))

    def __repr__(self) -> str:
        """String representation of the path."""
        return self._format_parts(self.parts)

    def _process_parts(
        self,
        parts: Iterable[Union[str, int, Sequence[Union[str, int]]]],
        allow_seq: bool = True,
    ) -> List[Union[str, int]]:
        """Helper method for converting initialization arguments to a list of parts."""
        split_parts: List[Union[str, int]] = []

        for part in parts:
            if isinstance(part, str):
                if part.isalpha():
                    # Part is trivially correct; skip complex parsing
                    split_parts.append(part)
                else:
                    split_parts.extend(self._parse_path_str(part))

            elif isinstance(part, int):
                split_parts.append(part)

            elif allow_seq and isinstance(part, Sequence):
                if isinstance(part, EPath):
                    # Part is already processed; skip complex parsing
                    split_parts.extend(part.parts)
                else:
                    split_parts.extend(self._process_parts(part, False))

            else:
                raise TypeError(
                    f"Invalid {self.__class__.__name__} part {part} of type '{type(part)}'"
                )

        return split_parts

    def _parse_path_str(self, part: str) -> Iterable[Union[str, int]]:
        """Convert a string path to a list of parts."""
        parsed_parts: List[Union[str, int]] = []

        remaining = part
        subpart_match = re.match(r"(?P<name>[a-zA-Z_][a-zA-Z0-9_]*)", remaining)
        if subpart_match is None:
            raise ValueError(f"Invalid {self.__class__.__name__} part '{part}'")

        parsed_parts.append(subpart_match["name"])
        remaining = remaining[subpart_match.end() :]

        while remaining:
            subpart_match = re.match(
                r"(?:(?:\[(?P<index>[0-9]+)\])|(?:\.(?P<name>[a-zA-Z_][a-zA-Z0-9_]*)))",
                remaining,
            )
            if subpart_match is None:
                raise ValueError(f"Invalid {self.__class__.__name__} part '{part}'")

            remaining = remaining[subpart_match.end() :]

            if (index := subpart_match["index"]) is not None:
                parsed_parts.append(int(index, 10))
            else:
                parsed_parts.append(subpart_match["name"])

        return parsed_parts

    @staticmethod
    def _format_parts(parts: Iterable[Union[str, int]]) -> str:
        """Format parts as a string path."""
        formatted_parts: List[str] = []

        for part in parts:
            if isinstance(part, int):
                formatted_parts.append(f"[{part}]")
            else:
                if not formatted_parts:
                    formatted_parts.append(part)
                else:
                    formatted_parts.append(f".{part}")

        return "".join(formatted_parts)


# Union of SVD path types
EPathUnion = Union[EPath, FEPath]

# Generic type variable constrained to being a specific SVD path type
EPathType = TypeVar("EPathType", EPath, FEPath)



#
# Copyright (c) 2023 Nordic Semiconductor ASA
#
# SPDX-License-Identifier: Apache-2.0
#

from abc import ABC
from typing import Any, Iterable, Union

from .path import EPath, EPathType, EPathUnion


class SvdError(Exception):
    """Base class for errors raised by the library."""

    ...


class SvdParseError(SvdError):
    """Raised when an error occurs during SVD parsing."""

    ...


class SvdDefinitionError(SvdError, ValueError):
    """Raised when unrecoverable errors occur due to an invalid definition in the SVD file."""

    def __init__(self, bindings: Iterable[Any], explanation: str):
        bindings_str = "\n".join(f"  * {b!r}" for b in bindings)
        super().__init__(f"Invalid SVD file element(s):\n{bindings_str}\n{explanation}")


class SvdMemoryError(SvdError, BufferError):
    """Raised when an invalid memory operation was attempted on a SVD element."""

    ...


class SvdPathError(SvdError):
    """Raised when trying to access a nonexistent/invalid SVD path."""

    def __init__(
        self, path: Union[str, EPathUnion], source: Any, explanation: str = ""
    ) -> None:
        formatted_explanation = "" if not explanation else f" ({explanation})"
        message = (
            f"{source!s} does not contain an element '{path}'{formatted_explanation}"
        )

        super().__init__(message)


class SvdIndexError(SvdPathError, IndexError):
    """Raised when given an invalid index in an array instance."""

    ...


class SvdKeyError(SvdPathError, KeyError):
    """Raised when given an invalid child element name in a struct or register instance. """

    ...



#
# Copyright (c) 2023 Nordic Semiconductor ASA
#
# SPDX-License-Identifier: Apache-2.0
#

from __future__ import annotations

from functools import partial
from typing import (
    Any,
    Callable,
    Dict,
    Iterator,
    List,
    Mapping,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
    overload,
)

import numpy as np
import numpy.ma as ma
from numpy.typing import ArrayLike, NDArray
from typing_extensions import Self

from .errors import SvdMemoryError

SIZE_TO_DTYPE: Mapping[int, np.dtype] = {
    1: np.dtype(np.uint8),
    2: np.dtype((np.dtype("<u2"), (np.uint8, 2))),
    4: np.dtype((np.dtype("<u4"), (np.uint8, 4))),
}


def _get_dtype_for_size(item_size: int) -> np.dtype:
    try:
        return SIZE_TO_DTYPE[item_size]
    except KeyError:
        raise SvdMemoryError(f"Unsupported item size: {item_size}")


IdxT = TypeVar("IdxT", int, slice)


class MemoryBlock:
    """
    A contiguous memory region at a given (offset, length).
    """

    class Builder:
        """
        Builder that can be used to construct a MemoryBlock in several steps.
        """

        def __init__(self) -> None:
            self._lazy_base_block: Optional[Callable[[], MemoryBlock]] = None
            self._offset: Optional[int] = None
            self._length: Optional[int] = None
            self._default_content: Optional[int] = None
            self._default_item_size: Optional[int] = None
            self._ops: List[Callable[[MemoryBlock], None]] = []

        def build(self) -> MemoryBlock:
            """
            Build the memory block based on the parameters set.

            :return: The built memory block.
            """
            if self._default_content is None or self._default_item_size is None:
                raise ValueError("Missing ")

            from_block: Optional[MemoryBlock] = (
                self._lazy_base_block() if self._lazy_base_block is not None else None
            )

            block = MemoryBlock(
                default_content=self._default_content,
                length=self._length,
                offset=self._offset,
                from_block=from_block,
            )

            for op in self._ops:
                op(block)

            return block

        def lazy_copy_from(self, lazy_block: Callable[[], MemoryBlock]) -> Self:
            """
            Use a different memory block as the base for this memory block.
            The lazy_block should be a callable that can be called in build() to get the base
            block.

            :param lazy_block: Callable that returns the base block.
            :return: The builder instance.
            """
            self._lazy_base_block = lazy_block
            return self

        def set_extent(self, offset: int, length: int) -> Self:
            """
            Set the offset and length of the memory block.
            This is required unless lazy_copy_from() is used.

            :param offset: Starting offset of the memory block.
            :param length: Length of the memory block, starting at the given offset.
            :return: The builder instance.
            """
            self._offset = offset
            self._length = length
            return self

        def set_default_content(self, default_content: int, item_size: int = 4) -> Self:
            """
            Set the default content (initial value) of the memory block.
            This is required.

            :param default_content: Default value.
            :param item_size: Size in bytes of default_content.
            :return: The builder instance.
            """
            self._default_content = default_content
            self._default_item_size = item_size
            return self

        def fill(self, start: int, end: int, content: int, item_size: int = 4) -> Self:
            """
            Fill the memory block address range [start, end) with a value.

            :param start: Start offset of the range to be filled.
            :param end: Exclusive end offset of the range to be filled.
            :param content: Value to fill with.
            :param item_size: Size in bytes of content.
            :return: The builder instance.
            """
            if start < end:
                self._ops.append(
                    partial(
                        MemoryBlock._fill,
                        start=start,
                        end=end,
                        value=content,
                        item_size=item_size,
                    )
                )
            return self

        def tile(self, start: int, end: int, times: int) -> Self:
            """
            Duplicate the values at the memory block address range [start, end) a number of times.
            The range is duplicated at the *times* positions following the range.

            :param start: Start offset of the range to be duplicated.
            :param end: Exclusive end offset of the range to be duplicated.
            :param times: Number of times to duplicate the range.
            :return: The builder instance.
            """
            if times > 1 and start < end:
                self._ops.append(
                    partial(MemoryBlock._tile, start=start, end=end, times=times)
                )
            return self

    def __init__(
        self,
        default_content: int,
        default_item_size: int = 4,
        offset: Optional[int] = None,
        length: Optional[int] = None,
        from_block: Optional[MemoryBlock] = None,
    ) -> None:
        """
        :param default_content:
        :param offset: Starting offset of the memory block. Required unless from_block is passed.
        :param length: Length in bytes of the memory block. Required unless from_block is passed.
        :param from_block: Memory block to use as the base for this memory block.
        """
        self._offset: int
        self._length: int

        default_dtype = SIZE_TO_DTYPE[default_item_size]

        if from_block is not None:
            # Shrink the offset and expand the length so that the base block is covered.
            if offset is not None:
                self._offset = min(offset, from_block._offset)
            else:
                self._offset = from_block._offset

            if length is not None:
                self._length = (
                    max(self._offset + length, from_block._offset + from_block._length)
                    - self._offset
                )
            else:
                self._length = from_block._length
        else:
            if offset is None or length is None:
                raise ValueError(
                    "offset and length are required when no from_block is given"
                )

            self._offset = offset
            self._length = length

        # Define the arrays used to represent the memory at [offset...offset + length]
        # To save some space, the offset is added on/subtracted in the API functions, so that the
        # array sizes are independent of the offset.
        # The address space defined in the SVD file may have discontinuities where there are no
        # elements. This is represented in the memory block by an address mask that masks out all
        # the unoccupied addresses.
        data = _numpy_full(
            self._length // default_item_size, default_content, dtype=default_dtype
        ).view(np.uint8)
        address_mask = np.ones_like(data, dtype=bool)
        self._array: ma.MaskedArray = ma.MaskedArray(
            data=data, mask=address_mask, dtype=np.uint8
        )
        self._written = np.zeros_like(data, dtype=np.uint8)

        if from_block is not None:
            # Copy the source block arrays into this block.
            dst_start = from_block._offset - offset if offset is not None else 0
            dst_end = dst_start + from_block._length
            self._array.mask[dst_start:dst_end] &= from_block._array.mask
            np.copyto(dst=self._array[dst_start:dst_end], src=from_block._array)
            np.copyto(dst=self._written[dst_start:dst_end], src=from_block._written)

    def is_written(self, idx: IdxT) -> bool:
        """:return: True if the given address has been explicitly written to, False otherwise."""
        translated_idx, dtype = self._translate_access(idx, item_size=1)
        return bool(self._written.view(dtype=dtype)[translated_idx].any())

    @overload
    def at(self, offset: int, item_size: int = 4) -> int:
        ...

    @overload
    def at(self, offset: slice, item_size: int = 4) -> ArrayLike:
        ...

    def at(
        self, offset: IdxT, item_size: int = 4
    ) -> Union[int, ArrayLike]:
        """
        Get the memory value at a given address offset.

        :param offset: Address offset or a slice of address offsets.
        :param item_size: Size of the value(s) to get.
        :return: The value(s) stored at the given offset(s).
        """
        translated_idx, dtype = self._translate_access(offset, item_size)
        return self._array.data.view(dtype=dtype)[translated_idx]

    def set_at(
        self,
        offset: IdxT,
        value: Union[int, ArrayLike],
        item_size: int = 4,
    ) -> None:
        """
        Set the memory value at a given index.

        :param offset: Address offset or a slice of address offsets.
        :param value: Value to write to the offset(s).
        :param item_size: Size of the value(s) to get.
        """
        translated_idx, dtype = self._translate_access(offset, item_size)
        self._array.data.view(dtype=dtype)[translated_idx] = value
        self._written.view(dtype=dtype)[translated_idx] = _one_bits(item_size)

    @overload
    def __getitem__(self, offset: int, /) -> int:
        ...

    @overload
    def __getitem__(self, offset: slice, /) -> ArrayLike:
        ...

    def __getitem__(self, offset: Union[int, slice], /) -> Union[int, ArrayLike]:
        """
        Get the memory value at a given index.
        Assumes a 4-byte item size; see MemoryBlock.set_at() to use a different item size.
        """
        return self.at(offset)

    def __setitem__(self, offset: IdxT, value: Union[int, ArrayLike], /) -> None:
        """
        Set the memory value at a given index.
        Assumes a 4-byte item size; see MemoryBlock.set_at() to use a different item size.
        """
        self.set_at(offset, value)

    def memory_iter(
        self,
        item_size: int = 4,
        with_offset: int = 0,
        written_only: bool = False,
    ) -> Iterator[Tuple[int, int]]:
        """
        Iterator over memory addresses and values contained in the block,
        in ascending address order.

        :param item_size: Size in bytes of each memory element.
        :param with_offset: Base address offset to use when iterating.
        :param native_byteorder: Cast from device byte order to native byte order.
        :param written_only: Yield only the adresses/values that have been explictly written to.
        :return: Iterator over (address, value) pairs describing the memory in the block.
        """
        if self._length % item_size != 0:
            raise ValueError(
                f"Memory block length {self._length} is not aligned to the item size "
                f"of {item_size}"
            )

        dtype = SIZE_TO_DTYPE[item_size]

        # Mask of locations that should be included in the result
        address_filter = ~self._array.mask
        if written_only:
            address_filter = np.logical_and(
                address_filter, self._written.view(dtype=bool)
            )

        # Array of addresses to return.
        # The internal array represents the memory at [offset...offset + length],
        # so we start the address range at offset plus the additional offset argument.
        # Note that we first apply the filter (which is at 1B granularity) and then cast to
        # the desired item size.
        address_start = self._offset + with_offset
        addresses = np.linspace(
            address_start,
            address_start + self._length,
            num=self._length,
            endpoint=False,
            dtype=int,
        )[address_filter][::item_size]

        # Array of values to return. These are also first filtered at a 1B granularity and the
        # cast to the desired item size
        values = self._array.data[address_filter].view(dtype)

        for address, value in zip(addresses, values):
            yield int(address), int(value)

    def __len__(self) -> int:
        """
        Length of the memory block.
        This may be different from the length provided in the constructor if a larger from_block
        was also given.
        """
        return len(self._array)

    def _translate_access(self, offset: IdxT, item_size: int) -> Tuple[IdxT, np.dtype]:
        """
        Translate index and size arguments given in the public API to indices in the internal
        numpy arrays.
        """
        translated_idx: IdxT
        dtype: np.dtype = SIZE_TO_DTYPE[item_size]

        if isinstance(offset, int):
            if offset % item_size != 0:
                raise ValueError(
                    f"Offset {offset} is not aligned to the item size of {item_size}"
                )

            translated_idx = (offset - self._offset) // item_size

        elif isinstance(offset, slice):
            if any(
                v is not None and v & item_size != 0
                for v in (offset.start, offset.stop, offset.step)
            ):
                raise ValueError(
                    f"Offsets {offset} are not aligned to the item size of {item_size}"
                )

            translated_idx = slice(
                (offset.start - self._offset) // item_size
                if offset.start is not None
                else None,
                (offset.stop - self._offset) // item_size
                if offset.stop is not None
                else None,
                (offset.step // item_size) if offset.step else None,
            )

        else:
            raise ValueError(f"Unsupported index: {offset}")

        return translated_idx, dtype

    def _tile(self, /, *, start: int, end: int, times: int) -> None:
        """Inline tile operation. See MemoryBlock.Builder.tile() for details"""
        length = end - start
        src_offset_start = start - self._offset
        src_offset_end = end - self._offset
        dst_offset_start = src_offset_start + length
        dst_offset_end = src_offset_start + length * times

        for array in (self._array.data, self._array.mask):
            array_src = array[src_offset_start:src_offset_end]
            array_dst = array[dst_offset_start:dst_offset_end].reshape(
                (times - 1, length)
            )
            array_dst[:] = array_src

    def _fill(self, /, *, start: int, end: int, value: int, item_size: int) -> None:
        """Inline fill operation. See MemoryBlock.Builder.fill() for details"""
        dtype = SIZE_TO_DTYPE[item_size]
        offset_start = start - self._offset
        offset_end = end - self._offset
        self._array.mask[offset_start:offset_end] = False
        data_dst = self._array.data[offset_start:offset_end].view(dtype)
        data_dst[:] = value


def _numpy_full(length: int, value: int, dtype: np.dtype) -> np.ndarray:
    """
    Replacement for np.full() with special handling for the zero case.
    numpy doesn't seem to handle numpy.full(..., fill_value=0) in any special way.
    Using np.zeros() for that case here provides a significant speedup for large arrays.
    """
    if value == 0:
        return np.zeros(length, dtype=dtype)
    else:
        return np.full(
            length,
            value,
            dtype=dtype,
        )


def _one_bits(item_size: int) -> int:
    """:return: bitfield of all ones with the given item size."""
    return 2 ** (8 * item_size) - 1



#
# Copyright (c) 2023 Nordic Semiconductor ASA
#
# SPDX-License-Identifier: Apache-2.0
#

"""
High level representation of a SVD device.
This representation does not aim to expose all the information contained in the SVD file,
but instead focuses on certain key features of the device description.

Since processing all the information contained in the SVD file can be computationally expensive,
many of the operations in this module lazily compute the data needed on first access.
"""

from __future__ import annotations

import dataclasses as dc
import math
import re
import typing
from abc import ABC, abstractmethod
from collections import defaultdict
from dataclasses import dataclass
from functools import cached_property
from types import MappingProxyType
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Collection,
    Dict,
    Generic,
    Iterable,
    Iterator,
    List,
    Literal,
    Mapping,
    NamedTuple,
    NoReturn,
    Optional,
    Protocol,
    Reversible,
    Sequence,
    Tuple,
    Type,
    TypedDict,
    TypeVar,
    Union,
    overload,
)

from typing_extensions import TypeGuard

from . import bindings
from ._device import (
    ChildIter,
    LazyFixedMapping,
    iter_merged,
    remove_registers,
    strip_suffix,
    svd_element_repr,
    topo_sort_derived_peripherals,
)
from .bindings import (
    Access,
    Cpu,
    Dimensions,
    FullRegisterProperties,
    ReadAction,
    RegisterProperties,
    WriteAction,
    WriteConstraint,
)
from .errors import (
    SvdDefinitionError,
    SvdIndexError,
    SvdKeyError,
    SvdMemoryError,
    SvdPathError,
)
from .memory_block import MemoryBlock
from .path import EPath, EPathType, FEPath
from .profiling import TimingReport, create_object_report, timed_method

if TYPE_CHECKING:
    from .parsing import Options


# Union of regular register types
RegisterUnion = Union["Array", "Register", "Struct"]

# Generic type variable contrained to being a regular register type
RegisterType = TypeVar("RegisterType", "Array", "Register", "Struct")

# Generic type variable of an array element type
ArrayMemberT = TypeVar("ArrayMemberT", "Register", "Struct")

# Union of flat register types
FlatRegisterUnion = Union["FlatRegister", "FlatStruct"]

# Generic type variable constrained to being a flat register type
FlatRegisterType = TypeVar("FlatRegisterType", "FlatRegister", "FlatStruct")

# Type variable constrained to either a regular or a flat register (but not a mix of regular/flat)
RegisterKindType = TypeVar("RegisterKindType", RegisterUnion, FlatRegisterUnion)

# Type variable constrained to either a regular or flat field
FieldKindType = TypeVar("FieldKindType", "Field", "FlatField")


class Device(Mapping[str, "Peripheral"]):
    """Representation of a SVD device."""

    @timed_method(max_times=1)
    def __init__(
        self,
        device: bindings.DeviceElement,
        options: Options,
        **kwargs: Any,
    ):
        self._device: bindings.DeviceElement = device
        self._options: Options = options
        self._reg_props: RegisterProperties = self._device.register_properties

        if self._device.address_unit_bits != 8:
            raise NotImplementedError(
                "the implementation assumes a byte-addressable device"
            )

        peripherals_unsorted: Dict[str, Peripheral] = {}

        # Initialize peripherals in topological order to ensure that base peripherals are
        # initialized before derived peripherals.
        for peripheral_element in topo_sort_derived_peripherals(device.peripherals):
            if peripheral_element.derived_from is not None:
                base_peripheral = peripherals_unsorted[peripheral_element.derived_from]
            else:
                base_peripheral = None

            if options is not None and options.skip_registers:
                remove_registers(peripheral_element, options.skip_registers)

            peripheral = Peripheral(
                peripheral_element,
                device=self,
                options=self._options,
                parent_reg_props=self._reg_props,
                base_peripheral=base_peripheral,
            )

            peripherals_unsorted[peripheral.name] = peripheral

        self._peripherals: Dict[str, Peripheral] = dict(
            sorted(peripherals_unsorted.items(), key=lambda kv: kv[1].base_address)
        )

        # Store the time spent parsing prior to calling this constructor
        self._time_parse = kwargs.get("time_parse", 0)

    @property
    def name(self) -> str:
        """Name of the device."""
        return self._device.name

    @property
    def series(self) -> Optional[str]:
        """Device series name."""
        return self._device.series

    @property
    def vendor_id(self) -> Optional[str]:
        """Device vendor ID."""
        return self._device.vendor_id

    @property
    def qualified_name(self) -> str:
        """Name of the device, including vendor and series."""
        return f"{self.vendor_id or ''}::{self.series or ''}::{self.name}"

    @property
    def cpu(self) -> Optional[Cpu]:
        """Device CPU information"""
        return self._device.cpu

    @property
    def address_unit_bits(self) -> int:
        """Number of data bits corresponding to an address."""
        return self._device.address_unit_bits

    @property
    def bus_bit_width(self) -> int:
        """Maximum data bits supported by the data bus in a single transfer."""
        return self._device.width

    @property
    def peripherals(self) -> Mapping[str, Peripheral]:
        """
        Map of peripherals in the device, indexed by name.
        The peripherals are sorted by ascending base address.
        """
        return MappingProxyType(self._peripherals)

    @property
    def _profiling_info(self) -> ProfilingInfo:
        """Profiling info for the device and peripherals."""
        return {
            "parse": float(self._time_parse),
            "device": create_object_report(self),
            "peripherals": {
                name: create_object_report(peripheral)
                for name, peripheral in self.items()
            },
        }

    def __getitem__(self, name: str) -> Peripheral:
        """
        :param name: Peripheral name.
        :raises SvdPathError: if the peripheral was not found.
        :return: Peripheral with the given name.
        """
        try:
            return self.peripherals[name]
        except LookupError as e:
            raise SvdKeyError(name, self, "peripheral not found") from e

    def __iter__(self) -> Iterator[str]:
        """:return: Iterator over the names of peripherals in the device."""
        return iter(self.peripherals)

    def __len__(self) -> int:
        """:return: Number of peripherals in the device."""
        return len(self.peripherals)

    def __repr__(self) -> str:
        """Short description of the device."""
        return svd_element_repr(self.__class__, self.qualified_name, length=len(self))


class ProfilingInfo(TypedDict):
    """Type of profiling information reported by the device."""

    parse: float
    device: TimingReport
    peripherals: Mapping[str, TimingReport]


class Peripheral(Mapping[str, RegisterUnion]):
    """
    Representation of a specific device peripheral.

    Internally, this class maintains a representation of a peripheral that is always guaranteed to
    be correct when compared to the allowable values prescribed by the SVD file the class was
    instantiated from. This representation starts off by having the default values defined within
    the SVD.
    """

    def __init__(
        self,
        element: bindings.PeripheralElement,
        device: Optional[Device],
        options: Options,
        parent_reg_props: bindings.RegisterProperties,
        base_peripheral: Optional[Peripheral] = None,
        new_base_address: Optional[int] = None,
    ):
        """
        :param element: SVD peripheral element.
        :param device: Parent Device object (may be None if this is a copy).
        :param base_reg_props: Register properties of the parent device.
        :param base_peripheral: Base peripheral that this peripheral is derived from, if any.
        :param new_base_address: Overridden base address of this peripheral, if any.
        """

        self._peripheral: bindings.PeripheralElement = element
        self._device: Optional[Device] = device
        self._options: Options = options
        self._base_peripheral: Optional[Peripheral] = base_peripheral
        self._base_address: int = (
            element.base_address if new_base_address is None else new_base_address
        )
        self._reg_props: bindings.RegisterProperties = (
            self._peripheral.get_register_properties(base_props=parent_reg_props)
        )

        # These dicts store every register associated with the peripheral.
        self._flat_registers: Dict[FEPath, FlatRegisterUnion] = {}
        self._dim_registers: Dict[EPath, RegisterUnion] = {}

    def copy_to(self, new_base_address: int) -> Peripheral:
        """
        Copy the peripheral to a new base address.

        :param new_base_address: Base address of the new peripheral.
        :returns: A copy of this peripheral at the new base address.
        """
        return Peripheral(
            element=self._peripheral,
            device=None,
            options=self._options,
            parent_reg_props=self._reg_props,
            base_peripheral=self,
            new_base_address=new_base_address,
        )

    @property
    def name(self) -> str:
        """Name of the peripheral."""
        return self._peripheral.name

    @property
    def base_address(self) -> int:
        """Base address of the peripheral."""
        return self._base_address

    @property
    def address_bounds(self) -> Tuple[int, int]:
        """Minimum and maximum address occupied by the registers in the peripheral."""
        bounds = self._register_info.address_bounds
        if bounds is None:
            raise SvdMemoryError(f"{self!s} has no registers")
        min_offset, max_offset = bounds
        return (self.base_address + min_offset, self.base_address + max_offset)

    @property
    def interrupts(self) -> Mapping[str, int]:
        """Interrupts associated with the peripheral. Mapping from interrupt name to value."""
        return {
            interrupt.name: interrupt.value for interrupt in self._peripheral.interrupts
        }

    @cached_property
    def registers(self) -> Mapping[str, RegisterUnion]:
        """Mapping of top-level registers in the peripheral, indexed by name."""
        return LazyFixedMapping(keys=self._specs.keys(), factory=self.__getitem__)

    @overload
    def register_iter(self, leaf_only: Literal[True]) -> Iterator[Register]:
        ...

    @overload
    def register_iter(self, leaf_only: bool = False) -> Iterator[RegisterUnion]:
        ...

    @timed_method(max_times=10)
    def register_iter(self, leaf_only: Any = False) -> Any:
        """
        Iterator over the registers in the peripheral in pre-order.
        Registers are ordered by increasing offset/address.

        :param flat: Do not yield individual registers in register arrays.
        :param leaf_only: Only yield registers at the bottom of the register tree.
        :return: Iterator over the registers in the peripheral.
        """
        return self._register_iter(EPath, leaf_only)

    @cached_property
    def flat_registers(self) -> Mapping[str, FlatRegisterUnion]:
        """
        Mapping of top-level flat registers in the peripheral, indexed by name.
        The
        """
        return LazyFixedMapping(
            keys=self._specs.keys(),
            factory=lambda n: self._get_or_create_register(FEPath(n)),
        )

    @overload
    def flat_register_iter(self, leaf_only: Literal[True]) -> Iterator[FlatRegister]:
        ...

    @overload
    def flat_register_iter(
        self, leaf_only: Literal[False]
    ) -> Iterator[FlatRegisterUnion]:
        ...

    @timed_method(max_times=10)
    def flat_register_iter(
        self, leaf_only: bool = False
    ) -> Iterator[FlatRegisterUnion]:
        """
        Iterator over the registers in the peripheral in pre-order.
        Registers are ordered by increasing offset/address.

        :param flat: Do not yield individual registers in register arrays.
        :param leaf_only: Only yield registers at the bottom of the register tree.
        :return: Iterator over the registers in the peripheral.
        """
        return self._register_iter(FEPath, leaf_only)

    def _register_iter(self, path_cls: Type, leaf_only: bool = False) -> Iterator:
        """Commmon register iteration implementation."""
        stack = [
            self._get_or_create_register(path_cls(name))
            for name in reversed(self._specs.keys())
        ]

        while stack:
            register = stack.pop()

            if register.leaf or not leaf_only:
                yield register

            if not register.leaf:
                stack.extend(reversed(register.child_iter()))

    @timed_method(max_times=10)
    def memory_iter(
        self,
        item_size: int = 1,
        absolute_addresses: bool = False,
        written_only: bool = False,
    ) -> Iterator[Tuple[int, int]]:
        """
        Get an iterator over the peripheral register contents.

        :param item_size: Byte granularity of the iterator.
        :param absolute_addresses: If True, use absolute instead of peripheral relative addresses.
        :param written_only: If true, only include those addresses that have been explicitly
        written to.
        :return: Iterator over the peripheral register contents.
        """
        address_offset = self.base_address if absolute_addresses else 0
        yield from self._memory_block.memory_iter(
            item_size, with_offset=address_offset, written_only=written_only
        )

    def __getitem__(self, path: Union[str, Sequence[Union[str, int]]]) -> RegisterUnion:
        """
        :param path: Name or path of the register.
        :return: Register instance.
        """
        return self._get_or_create_register(EPath(path))

    def __setitem__(self, name: str, content: int) -> None:
        """
        :param name: Name of the register to update.
        :param content: The raw register value to write to the specified register.
        """
        _register_set_content(self, name, content)

    def __iter__(self) -> Iterator[str]:
        """:return: Iterator over the names of top-level registers in the peripheral."""
        return iter(self._specs)

    def __len__(self) -> int:
        """:return: Number of top-level registers in the peripheral."""
        return len(self._specs)

    @overload
    def _get_or_create_register(self, path: EPath) -> RegisterUnion:
        ...

    @overload
    def _get_or_create_register(self, path: FEPath) -> FlatRegisterUnion:
        ...

    def _get_or_create_register(self, path: Any) -> Any:
        """
        Common method for accessing a register contained in the peripheral.
        If the register is accessed for the first time, it is first initialized.
        Otherwise, a cached register is returned.
        Note that if the requested register is below the top level, all the registers that are
        ancestors of the register are also initialized if needed.

        :param path: Path to the register.
        :return: The register instance at the given path.
        """
        if isinstance(path, EPath):
            return self._do_get_or_create_register(self._dim_registers, path)
        elif isinstance(path, FEPath):
            return self._do_get_or_create_register(self._flat_registers, path)
        else:
            raise ValueError(f"unrecognized path: {path}")

    @overload
    def _do_get_or_create_register(
        self, storage: Dict[EPath, RegisterUnion], path: EPath
    ) -> RegisterUnion:
        ...

    @overload
    def _do_get_or_create_register(
        self, storage: Dict[FEPath, FlatRegisterUnion], path: FEPath
    ) -> FlatRegisterUnion:
        ...

    def _do_get_or_create_register(self, storage: Dict, path: Any) -> Any:
        try:
            register = storage[path]
        except KeyError:
            ancestor_path = path.parent
            if ancestor_path is not None:
                # Build all the registers from the first available ancestor to the requested
                # register.
                while (
                    cached_register := storage.get(ancestor_path, None)
                ) is None and (parent := ancestor_path.parent) is not None:
                    ancestor_path = parent

                if cached_register is None:
                    register = self._create_register(ancestor_path)
                else:
                    register = cached_register

                for i in range(len(ancestor_path), len(path)):
                    register = self._create_register(path[: i + 1], register)
            else:
                register = self._create_register(path)

            storage[path] = register

        return register

    @overload
    def _create_register(
        self, path: EPath, parent: Optional[RegisterUnion] = None
    ) -> RegisterUnion:
        ...

    @overload
    def _create_register(
        self, path: FEPath, parent: Optional[FlatRegisterUnion] = None
    ) -> FlatRegisterUnion:
        ...

    def _create_register(
        self,
        path: Any,
        parent: Any = None,
    ) -> Any:
        if isinstance(path, EPath):
            return self._create_regular_register(path, parent)
        elif isinstance(path, FEPath):
            return self._create_flat_register(path, parent)
        else:
            raise ValueError(f"unrecognized path: {path}")

    def _create_regular_register(
        self, path: EPath, parent: Optional[RegisterUnion] = None
    ) -> RegisterUnion:
        if parent is None:
            try:
                spec = self._specs[path.stem]
            except KeyError:
                raise SvdPathError(path, self)

            instance_offset = 0

        elif isinstance(parent, Array):
            array_spec = parent._spec
            index = path.element_index
            if index is None:
                raise SvdIndexError(path, parent, "expected an array index path")

            try:
                array_offset = array_spec.dimensions.to_range()[index]
            except IndexError:
                raise SvdIndexError(
                    path,
                    parent,
                    f"index {index} is out of range for array with length "
                    f"{array_spec.dimensions.length}",
                )

            instance_offset = parent._instance_offset + array_offset
            spec = typing.cast(_RegisterSpec, parent._spec)

        elif isinstance(parent, Struct):
            try:
                spec = parent._spec.registers[path.stem]
            except KeyError:
                raise SvdKeyError(path.stem, parent)

            instance_offset = parent._instance_offset

        else:
            raise ValueError(f"Invalid parent register: {parent}")

        if is_array(spec) and path.element_index is None:
            return Array(
                spec=spec,
                peripheral=self,
                path=path,
                instance_offset=instance_offset,
            )
        elif is_struct(spec):
            return Struct(
                spec=spec,
                peripheral=self,
                path=path,
                instance_offset=instance_offset,
            )
        elif is_register(spec):
            return Register(
                spec=spec,
                peripheral=self,
                path=path,
                instance_offset=instance_offset,
            )
        else:
            raise ValueError(f"Element contains neither registers nor fields: {spec}")

    def _create_flat_register(
        self, path: FEPath, parent: Optional[FlatRegisterUnion] = None
    ) -> FlatRegisterUnion:
        """Create a flat register instance."""
        if parent is None:
            try:
                spec = self._specs[path.stem]
            except KeyError as e:
                raise SvdKeyError(path.stem, self) from e

            instance_offset = 0

        elif isinstance(parent, FlatStruct):
            try:
                spec = parent._spec.registers[path.stem]
            except KeyError as e:
                raise SvdKeyError(path.stem, parent) from e

            instance_offset = parent._instance_offset
        else:
            raise ValueError(f"Invalid parent register: {parent}")

        if is_struct(spec):
            return FlatStruct(
                spec=spec,
                peripheral=self,
                path=path,
                instance_offset=instance_offset,
            )
        elif is_register(spec):
            return FlatRegister(
                spec=spec,
                peripheral=self,
                path=path,
                instance_offset=instance_offset,
            )
        else:
            raise ValueError(f"Element contains neither registers nor fields: {spec}")

    @cached_property
    @timed_method(max_times=1)
    def _memory_block(self) -> MemoryBlock:
        """
        The memory block describing the values in the peripheral.
        This is computed the first time the property is accessed and cached for later.
        Initially this contains the reset values described in the SVD file.

        Note that accessing this property in a derived peripheral may also cause the memory blocks
        of base peripherals to be computed.
        """
        return self._register_info.memory_builder.build()

    @property
    def _specs(self) -> Mapping[str, _RegisterSpec]:
        """Mapping of register descriptions in the peripheral, indexed by name."""
        return self._register_info.register_specs

    @cached_property
    @timed_method(max_times=1)
    def _register_info(self) -> _ExtractedRegisterInfo:
        """
        Compute the descriptions of the registers contained in the peripheral, taking into
        account registers derived from the base peripheral, if any.
        """
        base_descriptions: Optional[Mapping[str, _RegisterSpec]] = None
        base_memory: Optional[Callable[[], MemoryBlock]] = None
        base_address_bounds: Optional[Tuple[int, int]] = None

        if self._base_peripheral is not None:
            base_peripheral: Peripheral = self._base_peripheral

            # If the register properties are equal, then it is possible to reuse all the immutable
            # properties from the base peripheral.
            if base_peripheral._reg_props == self._reg_props:
                base_descriptions = base_peripheral._specs
                base_memory = lambda: base_peripheral._memory_block
            # Otherwise, traverse the base registers again, because the difference in
            # register properties propagates down to the register elements.
            else:
                base_info = _extract_register_info(
                    base_peripheral._peripheral.registers,
                    options=self._options,
                    base_reg_props=self._reg_props,
                )
                base_descriptions = base_info.register_specs
                base_memory = lambda: base_info.memory_builder.build()

            base_address_bounds = base_peripheral._register_info.address_bounds

        try:
            info = _extract_register_info(
                self._peripheral.registers,
                options=self._options,
                base_reg_props=self._reg_props,
                base_specs=base_descriptions,
                base_memory=base_memory,
                base_address_bounds=base_address_bounds,
            )
        except _MissingDefaultResetValueError:
            raise SvdDefinitionError(
                [self._peripheral], "Missing default reset value for the peripheral."
            )

        return info

    def __repr__(self) -> str:
        """Short peripheral description."""
        return svd_element_repr(self.__class__, self.name, address=self.base_address)


def _register_set_content(container: Any, path: Union[int, str], content: int) -> None:
    """Common function for setting the content of a register."""
    try:
        container[path].content = content
    except AttributeError as e:
        raise SvdMemoryError(f"{container[path]!s} does not have content") from e


class _RegisterSpec(NamedTuple):
    """
    Immutable description of a SVD register/cluster element.
    This is separated from the register classes to optimize construction of derived peripherals.
    Since the description is not tied to a specific Peripheral instance, it can be
    instantiated once per register/cluster in the SVD document and reused for derived peripherals,
    as long as inherited properties are the same for the base and derived peripheral.
    """

    # Register name
    name: str
    # Lowest address offset contained within this element and any descendants
    offset_start: int
    # Highest address offset contained within this element and any descendants
    offset_end: int
    # Effective register properties, either inherited or specified on the element itself
    reg_props: bindings.RegisterProperties
    # Register dimensions
    dimensions: Optional[bindings.Dimensions]
    # Direct child registers of this element
    registers: Optional[Mapping[str, _RegisterSpec]]
    # Child fields of this element
    fields: Optional[Mapping[str, _FieldSpec]]
    # The SVD element itself
    element: Union[bindings.RegisterElement, bindings.ClusterElement]


class _CommonP(Protocol):
    """Protocol describing fields common to all register specs."""

    name: str
    offset_start: int
    offset_end: int
    reg_props: Union[RegisterProperties, FullRegisterProperties]
    dimensions: Optional[bindings.Dimensions]
    element: Union[bindings.RegisterElement, bindings.ClusterElement]


class _StructP(_CommonP):
    """
    Protocol describing fields specific to specs describing a struct.
    Used for type checking to avoid None and instance checks.
    """

    registers: Mapping[str, _RegisterSpec]
    reg_props: RegisterProperties
    element: bindings.ClusterElement


def is_struct(spec: _RegisterSpec) -> TypeGuard[_StructP]:
    """True if the SVD element describes a structure, i.e. contains other registers"""
    return spec.registers is not None


class _RegisterP(_CommonP):
    """
    Protocol describing fields specific to specs describing a register.
    Used for type checking to avoid None and instance checks.
    """

    reg_props: FullRegisterProperties
    fields: Mapping[str, _FieldSpec]
    element: bindings.RegisterElement


def is_register(spec: _RegisterSpec) -> TypeGuard[_RegisterP]:
    """True if the SVD element describes a register"""
    return spec.fields is not None


class _ArrayP(_CommonP):
    """
    Protocol describing fields specific to specs describing an array.
    Used for type checking to avoid None and instance checks.
    """

    dimensions: bindings.Dimensions


def is_array(spec: _RegisterSpec) -> TypeGuard[_ArrayP]:
    """True if the SVD element describes an array, i.e. has dimensions."""
    return spec.dimensions is not None


# Generic type variable describing a register spec variant.
# Register subclasses use one of the specific protocols this is constrained to for type checking.
_SpecT = TypeVar("_SpecT", _StructP, _RegisterP, _ArrayP)


class _RegisterNode(ABC, Generic[_SpecT, EPathType]):
    """Base class for all register level types"""

    __slots__ = [
        "_spec",
        "_peripheral",
        "_path",
        "_instance_offset",
    ]

    def __init__(
        self,
        spec: _SpecT,
        peripheral: Peripheral,
        path: EPathType,
        instance_offset: int = 0,
    ):
        """
        :param spec: Register description.
        :param peripheral: Parent peripheral.
        :param path: Path of the register.
        :param instance_offset: Address offset inherited from the parent register.
        """
        self._spec: _SpecT = spec
        self._peripheral: Peripheral = peripheral
        self._path: EPathType = path
        self._instance_offset: int = instance_offset

    @property
    def name(self) -> str:
        """Name of the register."""
        return self.path.name

    @property
    def path(self) -> EPathType:
        """Full path to the register."""
        return self._path

    @property
    def address(self) -> int:
        """Absolute address of the peripheral in memory"""
        return self._peripheral.base_address + self.offset

    @property
    def offset(self) -> int:
        """Address offset of the register, relative to the peripheral it is contained in"""
        return self._spec.offset_start + self._instance_offset

    @property
    @abstractmethod
    def leaf(self) -> bool:
        ...


class Array(_RegisterNode[_ArrayP, EPath], Sequence[ArrayMemberT]):
    """
    Container of regular Structs or Registers.

    In the hierarchy of regular register level types, an element with dimensions is
    expanded to an Array containing one object for each dimension index.
    The following SVD elements map to an Array object:
    * cluster element with dimensions -> Array[Struct]
    * register element with dimensions -> Array[Register]

    Arrays support the Sequence protocol, and can be used similarly to lists.
    """

    @overload
    def __getitem__(self, path: int, /) -> ArrayMemberT:
        ...

    @overload
    def __getitem__(self, path: slice, /) -> Sequence[ArrayMemberT]:
        ...

    @overload
    def __getitem__(self, path: Sequence[Union[int, str]], /) -> ArrayMemberT:
        ...

    def __getitem__(self, path: Any, /) -> Any:
        """Get the element at the given path."""
        if isinstance(path, slice):
            return [self[i] for i in range(*path.indices(len(self)))]

        return self._peripheral._get_or_create_register(self.path.join(path))

    def __setitem__(self, path: int, content: int, /) -> None:
        _register_set_content(self, path, content)

    def __iter__(self) -> Iterator[ArrayMemberT]:
        for i in range(len(self)):
            yield self[i]

    def __reversed__(self) -> Iterator[ArrayMemberT]:
        for i in reversed(range(len(self))):
            yield self[i]

    def __len__(self) -> int:
        """:return: Number of registers in the register array."""
        return self._spec.dimensions.length

    @property
    def leaf(self) -> bool:
        return False

    def child_iter(self) -> Reversible[ArrayMemberT]:
        return ChildIter(range(len(self)), self.__getitem__)

    def __repr__(self) -> str:
        """Short register description."""
        return svd_element_repr(
            self.__class__, str(self.path), address=self.offset, length=len(self)
        )


class _Struct(_RegisterNode[_StructP, EPathType], Mapping[str, RegisterKindType]):
    """Class implementing common struct functionality."""

    __slots__ = ["_registers"]

    def __init__(self, **kwargs: Any) -> None:
        """See parent class for a description of parameters."""
        super().__init__(**kwargs)

        self._registers: Optional[Mapping[str, RegisterKindType]] = None

    @property
    def leaf(self) -> bool:
        return False

    @property
    def registers(self) -> Mapping[str, RegisterKindType]:
        """:return A mapping of registers in the structure, ordered by ascending address."""
        if self._registers is None:
            self._registers = LazyFixedMapping(
                keys=iter(self), factory=self.__getitem__
            )

        return self._registers

    def __getitem__(
        self, path: Union[str, Sequence[Union[str, int]]]
    ) -> RegisterKindType:
        """
        :param index: Index of the register in the register array.
        :return: The instance of the specified register.
        """
        return self._peripheral._get_or_create_register(self.path.join(path))  # type: ignore

    def __iter__(self) -> Iterator[str]:
        """:return: Iterator over the names of registers in the register structure."""
        return iter(self._spec.registers)

    def __len__(self) -> int:
        """:return: Number of registers in the register structure"""
        return len(self._spec.registers)

    def child_iter(self) -> Reversible[RegisterKindType]:
        return ChildIter(self._spec.registers.keys(), self.__getitem__)

    def __repr__(self) -> str:
        """Short register description."""
        return svd_element_repr(self.__class__, str(self.path), address=self.offset)


class FlatStruct(_Struct[FEPath, FlatRegisterUnion]):
    """
    Flat structure representing a group of registers.
    These registers can be used to to read the properties defined on the SVD
    elements and inspect the structure of the SVD device.
    Flat structures are located at the offset specified on the corresponding element
    in the SVD file.

    A FlatStruct instance corresponds directly to a SVD cluster element,
    with or without dimensions.
    """

    @property
    def dimensions(self) -> Optional[Dimensions]:
        """Dimensions of the register, if any."""
        return self._spec.dimensions

    def __repr__(self) -> str:
        """Short struct description."""
        return svd_element_repr(
            self.__class__,
            str(self.path),
            address=self.offset,
            length=self.dimensions.length if self.dimensions is not None else None,
        )


class Struct(_Struct[EPath, RegisterUnion]):
    """
    Regular structure representing a group of registers.
    The address offset of a regular struct accounts for array element offsets of any
    parent elements.

    A Struct instance corresponds to either a SVD cluster element without dimensions,
    or a specific index of a cluster array element.
    """

    def __setitem__(self, path: str, content: int) -> None:
        _register_set_content(self, path, content)


class _Register(_RegisterNode[_RegisterP, EPathType], Mapping[str, FieldKindType]):
    """Commmon regular/flat register functionality."""

    __slots__ = ["_fields"]

    def __init__(self, **kwargs: Any) -> None:
        """See parent class for a description of parameters."""
        super().__init__(**kwargs)

        self._fields: Optional[Mapping[str, FieldKindType]] = None

    @property
    def leaf(self) -> bool:
        return True

    @property
    def bit_width(self) -> int:
        """Bit width of the register."""
        return self._spec.reg_props.size

    @property
    def access(self) -> Access:
        """Register access."""
        return self._spec.reg_props.access

    @property
    def reset_content(self) -> int:
        """Register reset value."""
        return self._spec.reg_props.reset_value

    @property
    def reset_mask(self) -> int:
        """Mask of bits in the register that are affected by a reset."""
        return self._spec.reg_props.reset_mask

    @property
    def write_action(self) -> WriteAction:
        """Side effect of writing the register"""
        return self._spec.element.modified_write_values

    @property
    def read_action(self) -> Optional[ReadAction]:
        """Side effect of reading from the register"""
        return self._spec.element.read_action

    @property
    def fields(self) -> Mapping[str, FieldKindType]:
        """Map of fields in the register, indexed by name"""
        if self._fields is None:
            self._fields = LazyFixedMapping(
                keys=self._spec.fields.keys(),
                factory=self._create_field,
            )

        return MappingProxyType(self._fields)

    def __getitem__(self, name: str) -> FieldKindType:
        """
        :param name: Field name.
        :return: The instance of the specified field.
        """
        try:
            return self.fields[name]
        except LookupError as e:
            raise SvdKeyError(
                name, self, explanation="no field matching the given path was found"
            ) from e

    def __iter__(self) -> Iterator[str]:
        """:return: Iterator over the field names in the register."""
        return iter(self._spec.fields)

    def __len__(self) -> int:
        """:return: Number of fields in the register."""
        return len(self._spec.fields)

    @abstractmethod
    def _create_field(self, name: str) -> FieldKindType:
        """Initialize the field with the given name."""
        ...


class FlatRegister(_Register[FEPath, "FlatField"]):
    """
    Flat register instance.
    These registers can be used to to read the properties defined on the SVD
    elements and inspect the structure of the SVD device.
    Flat registers ar located at the offset specified on the element in the SVD file.
    Unlike regular Register instances, flat registers do not contain a memory value.

    A FlatRegister instance corresponds directly to a SVD register element,
    with or without dimensions.
    """

    @property
    def dimensions(self) -> Optional[Dimensions]:
        """Dimensions of the register, if any."""
        return self._spec.dimensions

    def _create_field(self, name: str) -> FlatField:
        return FlatField(spec=self._spec.fields[name], register=self)

    def __repr__(self) -> str:
        return svd_element_repr(
            self.__class__,
            str(self.path),
            address=self.offset,
            length=self.dimensions.length if self.dimensions is not None else None,
        )


class Register(_Register[EPath, "Field"]):
    """
    Regular register instance.
    The address offset of a regular register accounts for array element offsets of any
    parent elements.
    Regular registers can be used to query or update the memory content of the peripheral.

    A Register instance corresponds to either a SVD register element without dimensions,
    or a specific index of a register array element.
    """

    @property
    def leaf(self) -> bool:
        return True

    @property
    def address_range(self) -> range:
        return range(self.address, self.address + self.bit_width // 8)

    @property
    def offset_range(self) -> range:
        """Range of addresses covered by the register."""
        return range(self.offset, self.offset + self.bit_width // 8)

    def __setitem__(self, key: str, content: Union[str, int]) -> None:
        """
        :param key: Either the bit offset of a field, or the field's name.
        :param content: A raw numeric value, or a field enumeration, to write
            to the selected register field.
        """
        self[key].content = content  # type: ignore

    @property
    def modified(self) -> bool:
        """True if the register contains a different value now than at reset."""
        return self.content != self.reset_content

    @property
    def written(self) -> bool:
        """True if the register content has been written."""
        return self._peripheral._memory_block.is_written(self.offset)

    @property
    def content(self) -> int:
        """Current value of the register."""
        content = self._peripheral._memory_block.at(self.offset, self.bit_width // 8)
        return int(content)

    @content.setter
    def content(self, new_content: int) -> None:
        """
        Set the value of the register.

        :param new_content: New value for the register.
        """
        self.set_content(new_content)

    def set_content(self, new_content: int, mask: Optional[int] = None) -> None:
        """
        Set the value of the register.

        :param new_content: New value for the register.
        :param mask: Mask of the bits to copy from the given value. If None, all bits are copied.
        """
        reg_width = self.bit_width

        if new_content > 0 and math.ceil(math.log2(new_content)) > reg_width:
            raise SvdMemoryError(
                f"Value {hex(new_content)} is too large for {reg_width}-bit register {self.path}."
            )

        for field in self.values():
            # Only check fields that are affected by the mask
            if mask is None or mask & field.mask:
                field_content = field._extract_content_from_register(new_content)
                if field_content not in field.allowed_values:
                    raise SvdMemoryError(
                        f"Value {hex(new_content)} is invalid for register {self.path}, as field "
                        f"{field.name} does not accept the value {hex(field_content)}."
                    )

        if mask is not None:
            # Update only the bits indicated by the mask
            new_content = (self.content & ~mask) | (new_content & mask)
        else:
            new_content = new_content

        self._peripheral._memory_block.set_at(
            self.offset, new_content, item_size=reg_width // 8
        )

    def unconstrain(self) -> None:
        """Remove all value constraints imposed on the register."""
        for field in self.values():
            field.unconstrain()

    def _create_field(self, name: str) -> Field:
        return Field(spec=self._spec.fields[name], register=self)

    def __repr__(self) -> str:
        """Short register description."""
        bool_props = ("modified",) if self.modified else ()

        return svd_element_repr(
            self.__class__,
            str(self.path),
            address=self.offset,
            content=self.content,
            bool_props=bool_props,
        )


class _FieldSpec(NamedTuple):
    """
    Class containing immutable data describing a SVD field element.
    This is separated from the Field class to optimize construction of derived peripherals.
    Since the description is not tied to a specific Peripheral instance, it can be
    instantiated once per field in the SVD document and reused for derived peripherals.
    """

    name: str
    bit_range: bindings.BitRange
    enums: Dict[str, int]
    allowed_values: Collection[int]
    element: bindings.FieldElement

    @classmethod
    def from_element(cls, element: bindings.FieldElement) -> _FieldSpec:
        """
        :param element: SVD field element binding object.
        :return: Description of the field.
        """

        name = element.name
        bit_range = element.bit_range

        # We do not support "do not care" bits, as by marking bits "x", see
        # SVD docs "/device/peripherals/peripheral/registers/.../enumeratedValue"
        if element.enumerated_values is not None:
            enums = {e.name: e.value for e in element.enumerated_values.enums}
        else:
            enums = {}

        allowed_values = set(enums.values()) if enums else range(2**bit_range.width)

        return cls(
            name=name,
            bit_range=bit_range,
            enums=enums,
            allowed_values=allowed_values,
            element=element,
        )


FieldParent = TypeVar("FieldParent", Register, FlatRegister)


class _Field(Generic[FieldParent]):
    """Common regular/flat field functionality."""

    __slots__ = ["_spec", "_register", "_allowed_values"]

    def __init__(
        self,
        spec: _FieldSpec,
        register: FieldParent,
    ):
        """
        Initialize the class attribute(s).

        :param description: Field description.
        :param register: Register to which the field belongs.
        """
        self._spec: _FieldSpec = spec
        self._register: FieldParent = register
        self._allowed_values: Collection[int] = spec.allowed_values

    @property
    def name(self) -> str:
        """Name of the field."""
        return self._spec.name

    @property
    def reset_content(self) -> int:
        """Default field value."""
        return self._extract_content_from_register(self._register.reset_content)

    @property
    def bit_offset(self) -> int:
        """Bit offset of the field."""
        return self._spec.bit_range.offset

    @property
    def bit_width(self) -> int:
        """Bit width of the field."""
        return self._spec.bit_range.width

    @property
    def mask(self) -> int:
        """Bitmask of the field."""
        return ((1 << self.bit_width) - 1) << self.bit_offset

    @property
    def access(self) -> Access:
        """Access property of the field."""
        if (field_access := self._spec.element.access) is not None:
            return field_access
        return self._register.access

    @property
    def write_action(self) -> WriteAction:
        """Side effect of writing to the field."""
        if (field_write_action := self._spec.element.modified_write_values) is not None:
            return field_write_action
        return self._register.write_action

    @property
    def read_action(self) -> Optional[ReadAction]:
        """Side effect of reading from the field."""
        if (field_read_action := self._spec.element.read_action) is not None:
            return field_read_action
        return self._register.read_action

    @property
    def allowed_values(self) -> Collection[int]:
        """
        Possible valid values for the bitfield.
        By default, the values allowed for the field are defined by the field enumeration
        values. If the field does not have enumerations, all values that fit within the
        field bit width are allowed.
        """
        return self._allowed_values

    @property
    def enums(self) -> Mapping[str, int]:
        """
        A mapping between field enumerations and their corresponding values.
        Field enumerations are values such as "Allowed" = 1, "NotAllowed" = 0
        and are defined by the device's SVD file. This may be an empty map,
        if enumerations are not applicable to the field.
        """
        return self._spec.enums

    def _extract_content_from_register(self, register_content: int) -> int:
        """
        Internal method for extracting the field value from the parent register value.

        :param register_value: Value of the parent register
        :return: Field value extracted based on the field bit range
        """
        return (register_content & self.mask) >> self.bit_offset


class FlatField(_Field):
    """
    Flat field instance.
    These fields can be used to to read the properties defined on the SVD
    elements and inspect the structure of the SVD device.
    Unlike regular Field instances, FlatFields do not contain a memory value.

    A FlatField instance corresponds directly to a SVD field element.
    """

    def __repr__(self) -> str:
        """Short field description."""
        return svd_element_repr(
            self.__class__,
            self.name,
            content_max_width=self.bit_width,
        )


class Field(_Field):
    """
    Regular field instance.
    Regular fields can be used to query or update the memory content of the peripheral
    in addition to reading the properties defined on the SVD elements.

    A Field instance corresponds directly to a SVD field element.
    """

    @property
    def content(self) -> int:
        """The value of the field."""
        return self._extract_content_from_register(self._register.content)

    @content.setter
    def content(self, new_content: Union[int, str]) -> None:
        """
        Set the value of the field.

        :param value: A numeric value, or the name of a field enumeration (if applicable), to
        write to the field.
        """
        if isinstance(new_content, int):
            val = self._trailing_zero_adjusted(new_content)

            if val not in self.allowed_values:
                raise ValueError(
                    f"{self!r} does not accept"
                    f" the bit value '{val}' ({hex(val)})."
                    " Are you sure you have an up to date .svd file?"
                )
            resolved_value = val
        elif isinstance(new_content, str):
            if new_content not in self.enums:
                raise ValueError(
                    f"{self!r} does not accept"
                    f" the enum '{new_content}'."
                    " Are you sure you have an up to date .svd file?"
                )
            resolved_value = self.enums[new_content]
        else:
            raise TypeError(
                f"Field does not accept write of '{new_content}' of type '{type(new_content)}'"
                " Permitted values types are 'str' (field enum) and 'int' (bit value)."
            )

        self._register.set_content(resolved_value << self.bit_offset, self.mask)

    @property
    def modified(self) -> bool:
        """True if the field contains a different value now than at reset."""
        return self.content != self.reset_content

    def unconstrain(self) -> None:
        """
        Remove restrictions on values that may be entered into this field. After this,
        the field will accept any value that can fit inside its bit width.
        """
        self._allowed_values = range(2**self.bit_width)

    def _trailing_zero_adjusted(self, content: int) -> int:
        """
        Internal method that checks and adjusts a given value for trailing zeroes if it exceeds
        the bit width of its field. Some values are simplest to encode as a full 32-bit value even
        though their field is comprised of less than 32 bits, such as an address.

        :param value: A numeric value to check against the field bits

        :return: Field value without any trailing zeroes
        """

        width_max = 2**self.bit_width

        if content > width_max:
            max_val = width_max - 1
            max_val_hex_len = len(f"{max_val:x}")
            hex_val = f"{content:0{8}x}"  # leading zeros, 8-byte max, in hex
            trailing = hex_val[max_val_hex_len:]  # Trailing zeros

            if int(trailing, 16) != 0:
                raise SvdMemoryError(f"Unexpected trailing value: {trailing}")

            cropped = hex_val[:max_val_hex_len]  # value w/o trailing
            adjusted = int(cropped, 16)

            if adjusted <= max_val:
                return adjusted

        return content

    def __repr__(self) -> str:
        """Short field description."""
        bool_props = ("modified",) if self.modified else ()

        return svd_element_repr(
            self.__class__,
            self.name,
            content=self.content,
            content_max_width=self.bit_width,
            bool_props=bool_props,
        )


class _ExtractedRegisterInfo(NamedTuple):
    """Container for register descriptions and reset values."""

    register_specs: Mapping[str, _RegisterSpec]
    memory_builder: MemoryBlock.Builder
    address_bounds: Optional[Tuple[int, int]]


def _extract_register_info(
    elements: Iterable[Union[bindings.RegisterElement, bindings.ClusterElement]],
    options: Options,
    base_reg_props: bindings.RegisterProperties,
    base_specs: Optional[Mapping[str, _RegisterSpec]] = None,
    base_memory: Optional[Callable[[], MemoryBlock]] = None,
    base_address_bounds: Optional[Tuple[int, int]] = None,
) -> _ExtractedRegisterInfo:
    """
    Extract register descriptions for the given SVD register level elements.
    The returned structure mirrors the structure of the SVD elements.
    Each level of the structure is internally sorted by ascending address.

    :param elements: Register level elements to process.
    :param options: Parsing options.
    :param base_reg_props: Register properties of the peripheral.
    :param base_descriptions: Register descriptions inherited from the base peripheral, if any.
    :param base_memory: Memory inherited from the base peripheral, if any.
    :return: Map of register descriptions, indexed by name.
    """
    memory_builder = MemoryBlock.Builder()

    if base_memory is not None:
        memory_builder.lazy_copy_from(base_memory)

    spec_list, min_address, max_address = _extract_register_descriptions_helper(
        memory=memory_builder,
        elements=elements,
        options=options,
        base_reg_props=base_reg_props,
    )

    specs = {d.name: d for d in spec_list}

    if base_specs is not None:
        # The register maps are each sorted internally, but need to be merged by address
        # to ensure sorted order in the combined map
        specs = dict(
            iter_merged(
                specs.items(),
                base_specs.items(),
                key=lambda kv: kv[1].offset_start,
            )
        )

    address_bounds: Optional[Tuple[int, int]]

    if spec_list:
        # Use the child address range if there is at least one child
        memory_builder.set_extent(offset=min_address, length=max_address - min_address)
        if base_address_bounds is not None:
            base_min, base_max = base_address_bounds
            address_bounds = (min(min_address, base_min), max(max_address, base_max))
        else:
            address_bounds = (min_address, max_address)
    else:
        address_bounds = base_address_bounds

    if base_reg_props.reset_value is None:
        # We don't have sufficient context to report this error in a nice way here,
        # so instead we raise this so that the caller can handle it.
        raise _MissingDefaultResetValueError()

    memory_builder.set_default_content(base_reg_props.reset_value)

    return _ExtractedRegisterInfo(specs, memory_builder, address_bounds)


class _MissingDefaultResetValueError(RuntimeError):
    """Signal that indicates a missing default reset value"""

    ...


class _ExtractHelperResult(NamedTuple):
    register_specs: List[_RegisterSpec]
    min_address: int
    max_address: int


def _extract_register_descriptions_helper(
    memory: MemoryBlock.Builder,
    elements: Iterable[Union[bindings.RegisterElement, bindings.ClusterElement]],
    options: Options,
    base_reg_props: bindings.RegisterProperties,
    base_address: int = 0,
) -> _ExtractHelperResult:
    """
    Helper that recursively extracts the names, addresses, register properties, dimensions,
    fields etc. of a collection of SVD register level elements.

    :param memory: Memory builder.
    :param elements: SVD register level elements.
    :param options: Parsing options.
    :param base_reg_props: Base address of the parent SVD element.
    :param base_address: Base address of the parent SVD element.

    :return: Extraction result.
    """

    specs: List[_RegisterSpec] = []
    min_address: int = 2**32
    max_address: int = 0

    for element in elements:
        # Remove suffixes used for elements with dimensions
        name = strip_suffix(element.name, "[%s]")
        reg_props = element.get_register_properties(base_props=base_reg_props)
        dim_props = element.dimensions
        address_offset = element.offset

        registers: Optional[Mapping[str, _RegisterSpec]] = None
        fields: Optional[Mapping[str, _FieldSpec]] = None

        if isinstance(element, bindings.RegisterElement):
            if not FullRegisterProperties.is_full(reg_props):
                raise SvdDefinitionError(
                    [element],
                    "Missing required register properties. "
                    f"Register has the following properties: {reg_props}",
                )

            # Register addresses are defined relative to the enclosing element
            if address_offset is not None:
                address_start = base_address + address_offset
            else:
                address_start = base_address

            size_bytes = reg_props.size // 8

            # Contiguous fill
            if dim_props is None or dim_props.step == size_bytes:
                length = dim_props.length if dim_props is not None else 1
                address_end = address_start + size_bytes * length
                memory.fill(
                    start=address_start,
                    end=address_end,
                    content=reg_props.reset_value,
                    item_size=size_bytes,
                )

            # Fill with gaps
            elif dim_props is not None and dim_props.step > size_bytes:
                memory.fill(
                    start=address_start,
                    end=address_start + size_bytes,
                    content=reg_props.reset_value,
                    item_size=size_bytes,
                )
                memory.tile(
                    start=address_start,
                    end=address_start + dim_props.step,
                    times=dim_props.length,
                )

            else:
                raise SvdDefinitionError(
                    element,
                    f"Step of 0x{dim_props.step:x} is less than the size of the "
                    f"array (0x{size_bytes})",
                )

            fields = _extract_field_descriptions(element.fields)

        else:  # ClusterElement
            if address_offset is not None:
                # By the SVD specification, cluster addresses are defined relative to the peripheral
                # base address, but some SVDs don't follow this rule.
                address_start = (
                    base_address
                    if not options.parent_relative_cluster_address
                    else base_address + address_offset
                )
            else:
                address_start = base_address

            (
                sub_specs,
                sub_min_address,
                sub_max_address,
            ) = _extract_register_descriptions_helper(
                memory=memory,
                elements=element.registers,
                options=options,
                base_reg_props=reg_props,
                base_address=address_start,
            )

            if sub_specs:
                registers = {d.name: d for d in sub_specs}

                if address_offset is None:
                    address_start = sub_min_address

                if dim_props is not None and dim_props.length > 1:
                    if dim_props.step < sub_max_address - address_start:
                        raise SvdDefinitionError(
                            element,
                            f"Step of 0x{dim_props.step:x} is less than the size required to "
                            f"cover all the child elements (0x{sub_max_address - address_start:x})",
                        )

                    address_end = address_start + dim_props.step * dim_props.length

                    # Copy memory from sub elements along the struct array dimension
                    memory.tile(
                        start=address_start,
                        end=address_start + dim_props.step,
                        times=dim_props.length,
                    )

                else:  # Not an array
                    address_end = sub_max_address

            else:  # Empty struct
                registers = {}
                address_end = address_start

        spec = _RegisterSpec(
            element=element,
            name=name,
            offset_start=address_start,
            offset_end=address_end,
            reg_props=reg_props,
            dimensions=dim_props,
            registers=registers,
            fields=fields,
        )

        specs.append(spec)
        min_address = min(min_address, address_start)
        max_address = max(max_address, address_end)

    sorted_result = sorted(specs, key=lambda r: r.offset_start)

    # Check that our structural assumptions hold.
    if not options.ignore_overlapping_structures:
        if len(sorted_result) > 1:
            for i in range(1, len(sorted_result)):
                r1 = sorted_result[i - 1]
                r2 = sorted_result[i]
                if r1.offset_end > r2.offset_start:
                    r1_str = f'"{r1.name}" ({r1.offset_start:#x}-{r1.offset_end:#x})'
                    r2_str = f'"{r2.name}" ({r2.offset_start:#x}-{r2.offset_end:#x})'
                    error_msg = f"Element addresses for {r1_str} and {r2_str} overlap"
                    raise SvdDefinitionError([r1.element, r2.element], error_msg)

    return _ExtractHelperResult(sorted_result, min_address, max_address)


def _extract_field_descriptions(
    elements: Iterable[bindings.FieldElement],
) -> Optional[Mapping[str, _FieldSpec]]:
    """
    Extract field descriptions for the given SVD field elements.
    The resulting mapping is internally sorted by ascending field bit offset.

    :param elements: Field elements to process.
    :return: Mapping of field descriptions, indexed by name.
    """

    field_specs = sorted(
        [_FieldSpec.from_element(field) for field in elements],
        key=lambda field: field.bit_range.offset,
    )

    fields = {spec.name: spec for spec in field_specs}

    return fields



#
# Copyright (c) 2023 Nordic Semiconductor ASA
#
# SPDX-License-Identifier: Apache-2.0
#

"""
Various internal functionality used by the bindings module.
"""

from __future__ import annotations

import enum
import inspect
import typing
from types import MappingProxyType
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    Iterable,
    Iterator,
    List,
    Literal,
    Mapping,
    Optional,
    Type,
    TypeVar,
    Union,
    overload,
)

from lxml import objectify
from typing_extensions import Self


class CaseInsensitiveStrEnum(enum.Enum):
    """String enum class that can be constructed from a case-insensitive string."""

    @classmethod
    def _missing_(cls, value: object) -> Optional[Self]:
        """Handler for string values with mismatched case."""
        if not isinstance(value, str):
            return None

        value_lower = value.lower()
        for member in cls:
            if member.value.lower() == value_lower:
                return member

        return None


def to_int(number: str) -> int:
    """
    Convert a string representation of an integer following the SVD format to its corresponding
    integer representation.

    :param number: String representation of the integer.

    :return: Decoded integer.
    """
    if number.startswith("0x"):
        return int(number, base=16)
    if number.startswith("#"):
        return int(number[1:], base=2)
    return int(number)


def to_bool(value: str) -> bool:
    """
    Convert a string representation of a boolean following the SVD format to its corresponding
    boolean representation.

    :param value: String representation of the boolean.

    :return: Decoded boolean.
    """
    if value in ("true", "1"):
        return True
    if value in ("false", "0"):
        return False
    raise ValueError(f"Invalid boolean value: {value}")


class SvdElement(objectify.ObjectifiedElement):
    """Base class for all the SVD element classes."""

    TAG: str

    def __repr__(self) -> str:
        """
        A more informative string representation than the default one from lxml.
        This is mostly useful for the exception tracebacks that occur on parsing errors.
        """
        return self._repr()

    def _repr(
        self,
        props: Mapping[Any, Any] = MappingProxyType({}),
    ) -> str:
        """
        Default repr() implementation for the binding classes.
        """

        props = dict(props)

        parent = self.getparent()
        if parent is not None:
            ancestors_str = f" in {parent!r}"
            try:
                child_index = parent.index(self)
            except Exception:
                child_index = None
        else:
            ancestors_str = ""
            child_index = None

        child_index_str = f"({child_index})" if child_index is not None else ""
        props_str = f" {props}" if props else ""
        self_repr = f"[{self.tag}{child_index_str}{props_str}]"

        return f"{self_repr}{ancestors_str}"


class SvdIntElement(objectify.IntElement):
    """
    Element containing an SVD integer value.
    This class uses a custom parser to convert the value to an integer.
    """

    def _init(self) -> None:
        self._setValueParser(to_int)


class _Self:
    ...


# Sentinel value used for element properties where the result class is equal to the class of the
# parent object. This is required since self-referential class members are not allowed.
SELF_CLASS = _Self()


class _Missing:
    ...


# Sentinel value used to indicate that a default value is missing.
MISSING = _Missing()


O = TypeVar("O", bound=objectify.ObjectifiedElement)
T = TypeVar("T")


class Elem(Generic[T]):
    """Data descriptor class used to access a XML element."""

    def __init__(
        self,
        name: str,
        element_class: Union[Type[objectify.ObjectifiedElement], _Self],
        /,
        *,
        default: Union[T, _Missing] = MISSING,
        default_factory: Union[Callable[[], T], _Missing] = MISSING,
    ) -> None:
        """
        Create a data descriptor object that extracts an element from an XML node.
        Only one of default or default_factory can be set.

        :param name: Name of the element.
        :param element_class: Class to use for the extracted element.
        :param default: Default value to return if the element is not found.
        :param default_factory: Callable that returns the default value to return if the element is
                                not found.
        """
        if default != MISSING and default_factory != MISSING:
            raise ValueError("Cannot set both default and default_factory")

        self.name: str = name
        self.element_class: Type[
            objectify.ObjectifiedElement
        ] = element_class  # type:ignore
        self.default: Union[T, _Missing] = default
        self.default_factory: Union[Callable[[], T], _Missing] = default_factory

    @overload
    def __get__(self, node: Literal[None], owner: Optional[Type] = None) -> Self:
        ...

    @overload
    def __get__(self, node: O, owner: Optional[Type] = None) -> T:
        ...

    def __get__(self, node: Optional[O], owner: Any = None) -> Union[T, Self]:
        """Get the element value from the given node."""
        if node is None:
            # If the node argument is None, we are being accessed through the class object.
            # In that case, return the descriptor itself.
            return self

        try:
            svd_obj = node.__getattr__(self.name)
        except AttributeError:
            if not isinstance(self.default_factory, _Missing):
                return self.default_factory()
            if not isinstance(self.default, _Missing):
                return self.default
            raise

        if issubclass(self.element_class, objectify.ObjectifiedDataElement):
            return svd_obj.pyval  # type: ignore
        else:
            return svd_obj  # type: ignore


class Attr(Generic[T]):
    """Data descriptor used to access a XML attribute."""

    def __init__(
        self,
        name: str,
        /,
        *,
        converter: Optional[Callable[[str], T]] = None,
        default: Union[T, _Missing] = MISSING,
        default_factory: Union[Callable[[], T], _Missing] = MISSING,
    ) -> None:
        """
        Create a data descriptor object that extracts an attribute from an XML node.
        Only one of default or default_factory can be set.

        :param name: Name of the attribute.
        :param converter: Optional callable that converts the attribute value from a string to another
                        type.
        :param default: Default value to return if the element is not found.
        :param default_factory: Callable that returns the default value to return if the element is
                                not found.
        """
        if default != MISSING and default_factory != MISSING:
            raise ValueError("Cannot set both default and default_factory")

        self.name: str = name
        self.converter: Optional[Callable[[str], T]] = converter
        self.default: Union[T, _Missing] = default
        self.default_factory: Union[Callable[[], T], _Missing] = default_factory

    @overload
    def __get__(self, node: Literal[None], owner: Optional[Type] = None) -> Self:
        ...

    @overload
    def __get__(self, node: O, owner: Optional[Type] = None) -> T:
        ...

    def __get__(self, node: Optional[O], owner: Any = None) -> Union[T, Self]:
        """Get the attribute value from the given node."""
        if node is None:
            # If the node argument is None, we are being accessed through the class object.
            # In that case, return the descriptor itself.
            return self

        value = node.get(self.name)

        if value is None:
            if not isinstance(self.default_factory, _Missing):
                return self.default_factory()
            if not isinstance(self.default, _Missing):
                return self.default
            raise AttributeError(f"Attribute {self.name} was not found")

        if self.converter is None:
            return value  # type: ignore

        try:
            return self.converter(value)
        except Exception as e:
            raise ValueError(f"Error converting attribute {self.name}") from e


C = TypeVar("C", bound=SvdElement)


class BindingRegistry:
    """Simple container for XML binding classes."""

    def __init__(self) -> None:
        self._element_classes: List[Type[SvdElement]] = []

    def add(
        self,
        element_class: Type[C],
        /,
    ) -> Type[C]:
        """
        Add a class to the binding registry.
        This is intended to be used as a class decorator.
        """
        elem_props: Dict[str, Elem] = getattr(element_class, "_xml_elem_props", {})
        attr_props: Dict[str, Attr] = getattr(element_class, "_xml_attr_props", {})

        for name, prop in inspect.getmembers(element_class):
            if isinstance(prop, Elem):
                if prop.element_class == SELF_CLASS:
                    prop.element_class = element_class
                elem_props[name] = prop

            elif isinstance(prop, Attr):
                attr_props[name] = prop

        setattr(element_class, "_xml_elem_props", elem_props)
        setattr(element_class, "_xml_attr_props", attr_props)

        self._element_classes.append(element_class)

        return element_class

    @property
    def bindings(self) -> List[Type[SvdElement]]:
        """Get the list of registered bindings."""
        return self._element_classes


def get_binding_elem_props(
    klass: Type[objectify.ObjectifiedElement],
) -> Mapping[str, Elem]:
    """Get the XML element properties of a binding class."""
    try:
        return klass._xml_elem_props  # type: ignore
    except AttributeError as e:
        raise ValueError(f"Class {klass} is not a binding") from e


def make_enum_wrapper(
    enum_cls: Type[CaseInsensitiveStrEnum],
) -> Type[SvdElement]:
    """
    Factory for creating lxml.objectify.ObjectifiedDataElement wrappers around
    CaseInsensitiveStrEnum subclasses.
    """

    class EnumWrapper(SvdElement, objectify.ObjectifiedDataElement):
        @property
        def pyval(self) -> CaseInsensitiveStrEnum:
            return enum_cls(self.text)

        def __repr__(self) -> str:
            props: Dict[str, Any]

            try:
                enum_value = self.pyval
                props = {"pyval": str(enum_value)}
            except Exception:
                props = {"text": self.text}

            return super()._repr(props=props)

    return EnumWrapper


def iter_element_children(
    element: Optional[objectify.ObjectifiedElement], *tags: str
) -> Iterable[objectify.ObjectifiedElement]:
    """
    Iterate over the children of an lxml element, optionally filtered by tag.
    If the element is None, an empty iterator is returned.
    """
    if element is None:
        return iter(())

    child_iter = element.iterchildren(*tags)  # type: ignore
    return typing.cast(Iterable[objectify.ObjectifiedElement], child_iter)



#
# Copyright (c) 2022 Nordic Semiconductor ASA
#
# SPDX-License-Identifier: Apache-2.0
#

from .bindings import (
    Access,
    ReadAction,
    Endian,
    SauAccess,
    AddressBlockUsage,
    Protection,
    EnumUsage,
    WriteAction,
    DataType,
    CpuName,
    Cpu,
    AddressBlock,
    SauRegion,
)
from .errors import (
    SvdError,
    SvdParseError,
    SvdDefinitionError,
    SvdMemoryError,
    SvdPathError,
    SvdIndexError,
    SvdKeyError,
)
from .parsing import (
    parse,
    Options,
)
from .path import FEPath, EPath
from .device import (
    Array,
    Field,
    FlatRegister,
    FlatRegisterUnion,
    FlatRegisterType,
    FlatStruct,
    FlatField,
    Device,
    Peripheral,
    Register,
    RegisterUnion,
    RegisterType,
    Struct,
)

import importlib.metadata

try:
    __version__ = importlib.metadata.version("svada")
except importlib.metadata.PackageNotFoundError:
    # Package is not installed
    import setuptools_scm
    __version__ = setuptools_scm.get_version(root="../..", relative_to=__file__)



#
# Copyright (c) 2022 Nordic Semiconductor ASA
#
# SPDX-License-Identifier: Apache-2.0
#

from __future__ import annotations

import dataclasses as dc
from collections import defaultdict
from dataclasses import dataclass
from pathlib import Path
from time import perf_counter_ns
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    List,
    Mapping,
    Optional,
    Sequence,
    Set,
    Tuple,
    Type,
    Union,
)

import lxml.etree as ET
from lxml import objectify

from . import bindings
from .device import Device
from .errors import SvdParseError

if TYPE_CHECKING:
    from ._bindings import SvdElement


@dataclass(frozen=True)
class Options:
    """Options to configure the SVD parsing behavior."""

    # Ignore overlapping clusters/registers described in the SVD file.
    # If set to False, an exception is raised when structures overlap.
    ignore_overlapping_structures: bool = False

    # Make the addressOffset in cluster elements be relative to the immediate parent
    # element instead of relative to the containing peripheral.
    parent_relative_cluster_address: bool = False

    # Cluster/register/field elements to remove from the XML document prior to parsing.
    # This can be used to remove outdated/deprecated elements from the device if they cause
    # issues with parsing.
    #
    # The value should be a dictionary mapping string peripheral name regex patterns to lists
    # containing the paths of elements to remove from the peripherals matching the pattern.
    # For example, passing the value
    # {"UART[0-9]": ["CONFIG.DEPRECATED"]}
    # would cause the element named "DEPRECATED" to be removed from the element named "CONFIG"
    # in peripherals whose name match "UART[0-9]" (UART0, UART1 etc.).
    #
    # Note: the element paths must match exactly the names used in the SVD document.
    # Note: no exception is raised if the paths given don't match anything in the SVD document.
    skip_registers: Mapping[str, Sequence[str]] = dc.field(
        default_factory=lambda: defaultdict(list)
    )


def parse(svd_path: Union[str, Path], options: Options = Options()) -> Device:
    """
    Parse a device described by a SVD file.

    :param svd_path: Path to the SVD file.
    :param options: Parsing options.

    :raises FileNotFoundError: If the SVD file does not exist.
    :raises SvdParseError: If an error occurred while parsing the SVD file.

    :return: Parsed `Device` representation of the SVD file.
    """

    t_parse_start = perf_counter_ns()

    svd_file = Path(svd_path)

    if not svd_file.is_file():
        raise FileNotFoundError(f"No such file: {svd_file.absolute()}")

    try:
        # Note: remove comments as otherwise these are present as nodes in the returned XML tree
        xml_parser = objectify.makeparser(remove_comments=True)
        class_lookup = _TwoLevelTagLookup(bindings.BINDINGS)
        xml_parser.set_element_class_lookup(class_lookup)

        with open(svd_file, "rb") as f:
            xml_device = objectify.parse(f, parser=xml_parser)

        t_parse = (perf_counter_ns() - t_parse_start) / 1_000_000
        device = Device(xml_device.getroot(), options=options, time_parse=t_parse)

    except Exception as e:
        raise SvdParseError(f"Error parsing SVD file {svd_file}") from e

    return device


class _TwoLevelTagLookup(ET.ElementNamespaceClassLookup):
    """
    XML element class lookup that uses two levels of tag names to map an XML element to a Python
    class. This two-level scheme is used to slightly optimize the time spent by the parser looking
    up looking up the class for an element (which is a sizeable portion of the time spent parsing).

    Element classes that can be uniquely identified by tag only are stored in the first level.
    This level uses the lxml ElementNamespaceClassLookup which is faster than the second level.
    The remaining element classes are assumed to be uniquely identified by a combination of
    the parent tag and the tag itself, and are stored in the second level.
    The second level uses the lxml PythonElementClassLookup which is slower.
    """

    def __init__(self, element_classes: List[Type[SvdElement]]):
        """
        :param element_classes: lxml element classes to add to the lookup table.
        """
        super().__init__()

        tag_classes: Dict[str, Set[type]] = defaultdict(set)
        two_tag_classes: Dict[Tuple[Optional[str], str], Set[type]] = defaultdict(set)

        for element_class in element_classes:
            tag = element_class.TAG
            tag_classes[tag].add(element_class)
            for prop in bindings.get_binding_elem_props(element_class).values():
                tag_classes[prop.name].add(prop.element_class)
                two_tag_classes[(tag, prop.name)].add(prop.element_class)

        one_tag: Set[str] = set()
        namespace = self.get_namespace(None)  # None is the empty namespace

        for tag, classes in tag_classes.items():
            if len(classes) == 1:
                # Add the class to the namespace
                # namespace is a decorator, so the syntax here is a little odd
                element_class = classes.pop()
                namespace(tag)(element_class)
                one_tag.add(tag)

        two_tag_lookup: Dict[Tuple[Optional[str], str], type] = {}

        for (parent_tag, field_name), classes in two_tag_classes.items():
            if field_name in one_tag:
                continue

            if len(classes) != 1:
                raise RuntimeError(
                    f"Multiple classes for ({parent_tag}, {field_name}): {classes}. "
                    "This should never happen, and likely indicates a bug in the way element "
                    "class lookup is implemented."
                )

            two_tag_lookup[(parent_tag, field_name)] = classes.pop()

        fallback_lookup = _SecondLevelTagLookup(two_tag_lookup)
        self.set_fallback(fallback_lookup)


class _SecondLevelTagLookup(ET.PythonElementClassLookup):
    """XML element class lookup table that uses two levels of tags to look up the class"""

    def __init__(
        self,
        lookup_table: Dict[
            Tuple[Optional[str], str], Type[objectify.ObjectifiedElement]
        ],
    ):
        """
        :param lookup_table: Lookup table mapping a tuple of (parent tag, tag) to an element class.
        """
        self._lookup_table = lookup_table

    def lookup(
        self, _document: Any, element: ET._Element
    ) -> Optional[Type[objectify.ObjectifiedElement]]:
        """Look up the Element class for the given XML element"""
        if (parent := element.getparent()) is not None:
            parent_tag = parent.tag
        else:
            parent_tag = None
        return self._lookup_table.get((parent_tag, element.tag))
